"""
Management of Heat
==================

.. versionadded:: 2017.7.0

:depends:   - heat Python module
:configuration: See :py:mod:`salt.modules.heat` for setup instructions.

The heat module is used to create, show, list and delete Heat staks.
Stack can be set as either absent or deploy.

.. code-block:: yaml

  heat.deployed:
    - name:
    - template: #Required
    - environment:
    - params: {}
    - poll: 5
    - rollback: False
    - timeout: 60

  heat.absent:
    - name:
    - poll: 5

mysql:
  heat.deployed:
    - template: salt://templates/mysql.heat.yaml
    - params:
      image: Debian 7
    - rollback: True

.. versionadded:: 2017.7.5,2018.3.1

    The spelling mistake in parameter `enviroment` was corrected to `environment`.
    The `enviroment` spelling mistake has been removed in Salt 3000.

"""

import logging

import salt.exceptions
import salt.utils.files
import salt.utils.json
import salt.utils.stringutils
import salt.utils.versions
import salt.utils.yaml

# pylint: disable=import-error
HAS_OSLO = False
try:
    from oslo_serialization import jsonutils

    HAS_OSLO = True
except ImportError:
    pass

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the mysql module is in __salt__
    """
    if HAS_OSLO:
        return "heat"
    return (
        False,
        "The heat state module cannot be loaded: "
        "the oslo_serialization python library is not available.",
    )


def _parse_template(tmpl_str):
    """
    Parsing template
    """
    tmpl_str = tmpl_str.strip()
    if tmpl_str.startswith("{"):
        tpl = salt.utils.json.loads(tmpl_str)
    else:
        try:
            tpl = salt.utils.yaml.safe_load(tmpl_str)
        except salt.utils.yaml.YAMLError as exc:
            raise ValueError(str(exc))
        else:
            if tpl is None:
                tpl = {}
    if not (
        "HeatTemplateFormatVersion" in tpl
        or "heat_template_version" in tpl
        or "AWSTemplateFormatVersion" in tpl
    ):
        raise ValueError("Template format version not found.")
    return tpl


def deployed(
    name,
    template=None,
    environment=None,
    params=None,
    poll=5,
    rollback=False,
    timeout=60,
    update=False,
    profile=None,
    **connection_args
):
    """
    Deploy stack with the specified properties

    name
        The name of the stack

    template
        File of template

    environment
        File of environment

    params
        Parameter dict used to create the stack

    poll
        Poll (in sec.) and report events until stack complete

    rollback
        Enable rollback on create failure

    timeout
        Stack creation timeout in minutes

    profile
        Profile to use

    .. versionadded:: 2017.7.5,2018.3.1

        The spelling mistake in parameter `enviroment` was corrected to `environment`.
        The `enviroment` spelling mistake has been removed in Salt 3000.

    """
    log.debug(
        "Deployed with(%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)",
        name,
        template,
        environment,
        params,
        poll,
        rollback,
        timeout,
        update,
        profile,
        connection_args,
    )
    ret = {"name": None, "comment": "", "changes": {}, "result": True}

    if not name:
        ret["result"] = False
        ret["comment"] = "Name ist not valid"
        return ret

    ret["name"] = (name,)

    existing_stack = __salt__["heat.show_stack"](name, profile=profile)

    if existing_stack["result"] and not update:
        ret["comment"] = "Stack {} is deployed".format(name)
        return ret
    if existing_stack["result"] and update:
        if template:
            template_tmp_file = salt.utils.files.mkstemp()
            tsfn, source_sum, comment_ = __salt__["file.get_managed"](
                name=template_tmp_file,
                template=None,
                source=template,
                source_hash=None,
                user=None,
                group=None,
                mode=None,
                saltenv="base",
                context=None,
                defaults=None,
                skip_verify=False,
                kwargs=None,
            )

            template_manage_result = __salt__["file.manage_file"](
                name=template_tmp_file,
                sfn=tsfn,
                ret=None,
                source=template,
                source_sum=source_sum,
                user=None,
                group=None,
                mode=None,
                saltenv="base",
                backup=None,
                makedirs=True,
                template=None,
                show_changes=False,
                contents=None,
                dir_mode=None,
            )

            if (template_manage_result["result"]) or (
                (__opts__["test"]) and (template_manage_result["result"] is not False)
            ):
                with salt.utils.files.fopen(template_tmp_file, "r") as tfp_:
                    tpl = salt.utils.stringutils.to_unicode(tfp_.read())
                    salt.utils.files.safe_rm(template_tmp_file)
                    try:
                        template_parse = _parse_template(tpl)
                        if "heat_template_version" in template_parse:
                            template_new = salt.utils.yaml.safe_dump(template_parse)
                        else:
                            template_new = jsonutils.dumps(
                                template_parse, indent=2, ensure_ascii=False
                            )
                        salt.utils.files.safe_rm(template_tmp_file)
                    except ValueError as ex:
                        ret["result"] = False
                        ret["comment"] = "Error parsing template {}".format(ex)
            else:
                ret["result"] = False
                ret["comment"] = "Can not open template: {} {}".format(
                    template, comment_
                )
        else:
            ret["result"] = False
            ret["comment"] = "Can not open template"
        if ret["result"] is True:
            template_stack = __salt__["heat.template_stack"](name=name, profile=profile)
            if not template_stack["result"]:
                ret["result"] = False
                ret["comment"] = template_stack["comment"]
        if ret["result"] is False:
            return ret

        try:
            checksum_template = __salt__["hashutil.digest"](template_new)
            checksum_stack = __salt__["hashutil.digest"](template_stack["template"])
        except salt.exceptions.CommandExecutionError as cmdexc:
            ret["result"] = False
            ret["comment"] = "{}".format(cmdexc)

        if ret["result"] is True:
            if checksum_template == checksum_stack:
                if __opts__["test"]:
                    ret["result"] = True
                    ret["comment"] = "Stack {} is deployed".format(name)
                    return ret
                else:
                    ret["result"] = False
                    ret["comment"] = "Templates have same checksum: {} {}".format(
                        checksum_template, checksum_stack
                    )
        if ret["result"] is False:
            return ret
        if __opts__["test"]:
            stack = {
                "result": None,
                "comment": "Stack {} is set to be updated".format(name),
            }
        else:
            stack = __salt__["heat.update_stack"](
                name=name,
                template_file=template,
                environment=environment,
                parameters=params,
                poll=poll,
                rollback=rollback,
                timeout=timeout,
                profile=profile,
            )
            ret["changes"]["stack_name"] = name
            ret["changes"]["comment"] = "Update stack"
    else:
        if __opts__["test"]:
            stack = {
                "result": None,
                "comment": "Stack {} is set to be created".format(name),
            }
        else:
            stack = __salt__["heat.create_stack"](
                name=name,
                template_file=template,
                environment=environment,
                parameters=params,
                poll=poll,
                rollback=rollback,
                timeout=timeout,
                profile=profile,
            )
            ret["changes"]["stack_name"] = name
            ret["changes"]["comment"] = "Create stack"
    ret["result"] = stack["result"]
    ret["comment"] = stack["comment"]

    return ret


def absent(name, poll=5, timeout=60, profile=None):
    """
    Ensure that the named stack is absent

    name
        The name of the stack to remove

    poll
        Poll(in sec.) and report events until stack complete

    timeout
        Stack creation timeout in minutes

    profile
        Profile to use

    """
    log.debug("Absent with(%s, %s %s)", name, poll, profile)
    ret = {"name": None, "comment": "", "changes": {}, "result": True}
    if not name:
        ret["result"] = False
        ret["comment"] = "Name ist not valid"
        return ret

    ret["name"] = (name,)

    existing_stack = __salt__["heat.show_stack"](name, profile=profile)

    if not existing_stack["result"]:
        ret["result"] = True
        ret["comment"] = "Stack not exist"
        return ret
    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "Stack {} is set to be removed".format(name)
        return ret

    stack = __salt__["heat.delete_stack"](
        name=name, poll=poll, timeout=timeout, profile=profile
    )

    ret["result"] = stack["result"]
    ret["comment"] = stack["comment"]
    ret["changes"]["stack_name"] = name
    ret["changes"]["comment"] = "Delete stack"
    return ret



"""
Functions for creating and working with job IDs
"""

import datetime
import hashlib
import os
from calendar import month_abbr as months

import salt.utils.stringutils

LAST_JID_DATETIME = None


def _utc_now():
    """
    Helper method so tests do not have to patch the built-in method.
    """
    return datetime.datetime.utcnow()


def gen_jid(opts):
    """
    Generate a jid
    """
    global LAST_JID_DATETIME  # pylint: disable=global-statement

    jid_dt = _utc_now()
    if not opts.get("unique_jid", False):
        return "{:%Y%m%d%H%M%S%f}".format(jid_dt)
    if LAST_JID_DATETIME and LAST_JID_DATETIME >= jid_dt:
        jid_dt = LAST_JID_DATETIME + datetime.timedelta(microseconds=1)
    LAST_JID_DATETIME = jid_dt
    return "{:%Y%m%d%H%M%S%f}_{}".format(jid_dt, os.getpid())


def is_jid(jid):
    """
    Returns True if the passed in value is a job id
    """
    if not isinstance(jid, str):
        return False
    if len(jid) != 20 and (len(jid) <= 21 or jid[20] != "_"):
        return False
    try:
        int(jid[:20])
        return True
    except ValueError:
        return False


def jid_to_time(jid):
    """
    Convert a salt job id into the time when the job was invoked
    """
    jid = str(jid)
    if len(jid) != 20 and (len(jid) <= 21 or jid[20] != "_"):
        return ""
    year = jid[:4]
    month = jid[4:6]
    day = jid[6:8]
    hour = jid[8:10]
    minute = jid[10:12]
    second = jid[12:14]
    micro = jid[14:20]

    ret = "{}, {} {} {}:{}:{}.{}".format(
        year, months[int(month)], day, hour, minute, second, micro
    )
    return ret


def format_job_instance(job):
    """
    Format the job instance correctly
    """
    ret = {
        "Function": job.get("fun", "unknown-function"),
        "Arguments": list(job.get("arg", [])),
        # unlikely but safeguard from invalid returns
        "Target": job.get("tgt", "unknown-target"),
        "Target-type": job.get("tgt_type", "list"),
        "User": job.get("user", "root"),
    }

    if "metadata" in job:
        ret["Metadata"] = job.get("metadata", {})
    else:
        if "kwargs" in job:
            if "metadata" in job["kwargs"]:
                ret["Metadata"] = job["kwargs"].get("metadata", {})
    return ret


def format_jid_instance(jid, job):
    """
    Format the jid correctly
    """
    ret = format_job_instance(job)
    ret.update({"StartTime": jid_to_time(jid)})
    return ret


def format_jid_instance_ext(jid, job):
    """
    Format the jid correctly with jid included
    """
    ret = format_job_instance(job)
    ret.update({"JID": jid, "StartTime": jid_to_time(jid)})
    return ret


def jid_dir(jid, job_dir=None, hash_type="sha256"):
    """
    Return the jid_dir for the given job id
    """
    if not isinstance(jid, str):
        jid = str(jid)
    jhash = getattr(hashlib, hash_type)(
        salt.utils.stringutils.to_bytes(jid)
    ).hexdigest()

    parts = []
    if job_dir is not None:
        parts.append(job_dir)
    parts.extend([jhash[:2], jhash[2:]])
    return os.path.join(*parts)



"""
Tests for the Openstack Cloud Provider
"""

import logging

import pytest
from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest
from tests.support.case import ModuleCase
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.unit import skipIf

log = logging.getLogger(__name__)

try:
    import keystoneclient  # pylint: disable=import-error,unused-import
    from libcloud.common.openstack_identity import OpenStackIdentity_3_0_Connection
    from libcloud.common.openstack_identity import OpenStackIdentityTokenScope

    HAS_KEYSTONE = True
except ImportError:
    HAS_KEYSTONE = False

try:
    import shade  # pylint: disable=unused-import

    HAS_SHADE = True
except ImportError:
    HAS_SHADE = False


@skipIf(
    not HAS_KEYSTONE,
    "Please install keystoneclient and a keystone server before running"
    "openstack integration tests.",
)
class OpenstackTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the keystone state
    """

    endpoint = "http://localhost:35357/v2.0"
    token = "administrator"

    @pytest.mark.destructive_test
    def test_aaa_setup_keystone_endpoint(self):
        ret = self.run_state(
            "keystone.service_present",
            name="keystone",
            description="OpenStack Identity",
            service_type="identity",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(
            ret["keystone_|-keystone_|-keystone_|-service_present"]["result"]
        )

        ret = self.run_state(
            "keystone.endpoint_present",
            name="keystone",
            region="RegionOne",
            publicurl="http://localhost:5000/v2.0",
            internalurl="http://localhost:5000/v2.0",
            adminurl="http://localhost:35357/v2.0",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(
            ret["keystone_|-keystone_|-keystone_|-endpoint_present"]["result"]
        )

        ret = self.run_state(
            "keystone.tenant_present",
            name="admin",
            description="Admin Project",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-tenant_present"]["result"])

        ret = self.run_state(
            "keystone.tenant_present",
            name="demo",
            description="Demo Project",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-demo_|-demo_|-tenant_present"]["result"])

        ret = self.run_state(
            "keystone.role_present",
            name="admin",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-role_present"]["result"])

        ret = self.run_state(
            "keystone.role_present",
            name="user",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-user_|-user_|-role_present"]["result"])

        ret = self.run_state(
            "keystone.user_present",
            name="admin",
            email="admin@example.com",
            password="adminpass",
            tenant="admin",
            roles={"admin": ["admin"]},
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-user_present"]["result"])

        ret = self.run_state(
            "keystone.user_present",
            name="demo",
            email="demo@example.com",
            password="demopass",
            tenant="demo",
            roles={"demo": ["user"]},
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-demo_|-demo_|-user_present"]["result"])

    @pytest.mark.destructive_test
    def test_zzz_teardown_keystone_endpoint(self):
        ret = self.run_state(
            "keystone.user_absent",
            name="admin",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-user_absent"]["result"])

        ret = self.run_state(
            "keystone.user_absent",
            name="demo",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-demo_|-demo_|-user_absent"]["result"])

        ret = self.run_state(
            "keystone.role_absent",
            name="admin",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-role_absent"]["result"])

        ret = self.run_state(
            "keystone.role_absent",
            name="user",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-user_|-user_|-role_absent"]["result"])

        ret = self.run_state(
            "keystone.tenant_absent",
            name="admin",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-admin_|-admin_|-tenant_absent"]["result"])

        ret = self.run_state(
            "keystone.tenant_absent",
            name="demo",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(ret["keystone_|-demo_|-demo_|-tenant_absent"]["result"])

        ret = self.run_state(
            "keystone.service_absent",
            name="keystone",
            connection_endpoint=self.endpoint,
            connection_token=self.token,
        )
        self.assertTrue(
            ret["keystone_|-keystone_|-keystone_|-service_absent"]["result"]
        )

    @pytest.mark.destructive_test
    def test_libcloud_auth_v3(self):
        driver = OpenStackIdentity_3_0_Connection(
            auth_url="http://localhost:5000",
            user_id="admin",
            key="adminpass",
            token_scope=OpenStackIdentityTokenScope.PROJECT,
            domain_name="Default",
            tenant_name="admin",
        )
        driver.authenticate()
        self.assertTrue(driver.auth_token)


@skipIf(not HAS_SHADE, "openstack driver requires `shade`")
class RackspaceTest(CloudTest):
    """
    Integration tests for the Rackspace cloud provider using the Openstack driver
    """

    PROVIDER = "openstack"
    REQUIRED_PROVIDER_CONFIG_ITEMS = ("auth", "cloud", "region_name")

    def test_instance(self):
        """
        Test creating an instance on rackspace with the openstack driver
        """
        # check if instance with salt installed returned
        ret_val = self.run_cloud(
            "-p rackspace-test {}".format(self.instance_name), timeout=TIMEOUT
        )
        self.assertInstanceExists(ret_val)

        self.assertDestroyInstance()



import salt.modules.boto_elbv2 as boto_elbv2
from tests.support.mock import patch
from tests.support.unit import TestCase


class BotoElbv2TestCase(TestCase):
    """
    TestCase for salt.modules.boto_elbv2 module
    """

    @patch("salt.utils.versions.check_boto_reqs")
    @patch("salt.utils.boto3mod.assign_funcs")
    def test___virtual__has_boto_reqs_true(self, mock_assign_funcs, mock_boto_reqs):
        mock_boto_reqs.return_value = True
        mock_assign_funcs.return_value = False
        result = boto_elbv2.__virtual__()
        self.assertEqual(result, True)

    @patch("salt.utils.versions.check_boto_reqs")
    def test___virtual__has_boto_reqs_false(self, mock_boto_reqs):
        mock_boto_reqs.return_value = False
        result = boto_elbv2.__virtual__()
        self.assertEqual(result, False)



"""
Support for Config Server Firewall (CSF)
========================================
:maintainer: Mostafa Hussein <mostafa.hussein91@gmail.com>
:maturity: new
:platform: Linux
"""

import re

import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltInvocationError


def __virtual__():
    """
    Only load if csf exists on the system
    """
    if salt.utils.path.which("csf") is None:
        return (False, "The csf execution module cannot be loaded: csf unavailable.")
    else:
        return True


def _temp_exists(method, ip):
    """
    Checks if the ip exists as a temporary rule based
    on the method supplied, (tempallow, tempdeny).
    """
    _type = method.replace("temp", "").upper()
    cmd = (
        "csf -t | awk -v code=1 -v type=_type -v ip=ip '$1==type && $2==ip {{code=0}}"
        " END {{exit code}}'".format(_type=_type, ip=ip)
    )
    exists = __salt__["cmd.run_all"](cmd)
    return not bool(exists["retcode"])


def _exists_with_port(method, rule):
    path = "/etc/csf/csf.{}".format(method)
    return __salt__["file.contains"](path, rule)


def exists(
    method,
    ip,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="d",
    ttl=None,
    comment="",
):
    """
    Returns true a rule for the ip already exists
    based on the method supplied. Returns false if
    not found.

    CLI Example:

    .. code-block:: bash

        salt '*' csf.exists allow 1.2.3.4
        salt '*' csf.exists tempdeny 1.2.3.4
    """
    if method.startswith("temp"):
        return _temp_exists(method, ip)
    if port:
        rule = _build_port_rule(
            ip, port, proto, direction, port_origin, ip_origin, comment
        )
        return _exists_with_port(method, rule)
    exists = __salt__["cmd.run_all"]("egrep ^'{} +' /etc/csf/csf.{}".format(ip, method))
    return not bool(exists["retcode"])


def __csf_cmd(cmd):
    """
    Execute csf command
    """
    csf_cmd = "{} {}".format(salt.utils.path.which("csf"), cmd)
    out = __salt__["cmd.run_all"](csf_cmd)

    if out["retcode"] != 0:
        if not out["stderr"]:
            ret = out["stdout"]
        else:
            ret = out["stderr"]
        raise CommandExecutionError("csf failed: {}".format(ret))
    else:
        ret = out["stdout"]
    return ret


def _status_csf():
    """
    Return True if csf is running otherwise return False
    """
    cmd = "test -e /etc/csf/csf.disable"
    out = __salt__["cmd.run_all"](cmd)
    return bool(out["retcode"])


def _get_opt(method):
    """
    Returns the cmd option based on a long form argument.
    """
    opts = {
        "allow": "-a",
        "deny": "-d",
        "unallow": "-ar",
        "undeny": "-dr",
        "tempallow": "-ta",
        "tempdeny": "-td",
        "temprm": "-tr",
    }
    return opts[method]


def _build_args(method, ip, comment):
    """
    Returns the cmd args for csf basic allow/deny commands.
    """
    opt = _get_opt(method)
    args = "{} {}".format(opt, ip)
    if comment:
        args += " {}".format(comment)
    return args


def _access_rule(
    method,
    ip=None,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="d",
    comment="",
):
    """
    Handles the cmd execution for allow and deny commands.
    """
    if _status_csf():
        if ip is None:
            return {"error": "You must supply an ip address or CIDR."}
        if port is None:
            args = _build_args(method, ip, comment)
            return __csf_cmd(args)
        else:
            if method not in ["allow", "deny"]:
                return {
                    "error": (
                        "Only allow and deny rules are allowed when specifying a port."
                    )
                }
            return _access_rule_with_port(
                method=method,
                ip=ip,
                port=port,
                proto=proto,
                direction=direction,
                port_origin=port_origin,
                ip_origin=ip_origin,
                comment=comment,
            )


def _build_port_rule(ip, port, proto, direction, port_origin, ip_origin, comment):
    kwargs = {
        "ip": ip,
        "port": port,
        "proto": proto,
        "direction": direction,
        "port_origin": port_origin,
        "ip_origin": ip_origin,
    }
    rule = "{proto}|{direction}|{port_origin}={port}|{ip_origin}={ip}".format(**kwargs)
    if comment:
        rule += " #{}".format(comment)

    return rule


def _remove_access_rule_with_port(
    method,
    ip,
    port,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="d",
    ttl=None,
):

    rule = _build_port_rule(
        ip,
        port=port,
        proto=proto,
        direction=direction,
        port_origin=port_origin,
        ip_origin=ip_origin,
        comment="",
    )

    rule = rule.replace("|", "[|]")
    rule = rule.replace(".", "[.]")
    result = __salt__["file.replace"](
        "/etc/csf/csf.{}".format(method),
        pattern="^{}(( +)?\\#.*)?$\n".format(rule),  # pylint: disable=W1401
        repl="",
    )

    return result


def _csf_to_list(option):
    """
    Extract comma-separated values from a csf.conf
    option and return a list.
    """
    result = []
    line = get_option(option)
    if line:
        csv = line.split("=")[1].replace(" ", "").replace('"', "")
        result = csv.split(",")
    return result


def split_option(option):
    return re.split(r"(?: +)?\=(?: +)?", option)


def get_option(option):
    pattern = r'^{}(\ +)?\=(\ +)?".*"$'.format(option)  # pylint: disable=W1401
    grep = __salt__["file.grep"]("/etc/csf/csf.conf", pattern, "-E")
    if "stdout" in grep and grep["stdout"]:
        line = grep["stdout"]
        return line
    return None


def set_option(option, value):
    current_option = get_option(option)
    if not current_option:
        return {"error": "No such option exists in csf.conf"}
    result = __salt__["file.replace"](
        "/etc/csf/csf.conf",
        pattern=r'^{}(\ +)?\=(\ +)?".*"'.format(option),  # pylint: disable=W1401
        repl='{} = "{}"'.format(option, value),
    )

    return result


def get_skipped_nics(ipv6=False):
    if ipv6:
        option = "ETH6_DEVICE_SKIP"
    else:
        option = "ETH_DEVICE_SKIP"

    skipped_nics = _csf_to_list(option)
    return skipped_nics


def skip_nic(nic, ipv6=False):
    nics = get_skipped_nics(ipv6=ipv6)
    nics.append(nic)
    return skip_nics(nics, ipv6)


def skip_nics(nics, ipv6=False):
    if ipv6:
        ipv6 = "6"
    else:
        ipv6 = ""
    nics_csv = ",".join(map(str, nics))
    result = __salt__["file.replace"](
        "/etc/csf/csf.conf",
        # pylint: disable=anomalous-backslash-in-string
        pattern=r'^ETH{}_DEVICE_SKIP(\ +)?\=(\ +)?".*"'.format(ipv6),
        # pylint: enable=anomalous-backslash-in-string
        repl='ETH{}_DEVICE_SKIP = "{}"'.format(ipv6, nics_csv),
    )

    return result


def _access_rule_with_port(
    method,
    ip,
    port,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="d",
    ttl=None,
    comment="",
):

    results = {}
    if direction == "both":
        directions = ["in", "out"]
    else:
        directions = [direction]
    for direction in directions:
        _exists = exists(
            method,
            ip,
            port=port,
            proto=proto,
            direction=direction,
            port_origin=port_origin,
            ip_origin=ip_origin,
            ttl=ttl,
            comment=comment,
        )
        if not _exists:
            rule = _build_port_rule(
                ip,
                port=port,
                proto=proto,
                direction=direction,
                port_origin=port_origin,
                ip_origin=ip_origin,
                comment=comment,
            )
            path = "/etc/csf/csf.{}".format(method)
            results[direction] = __salt__["file.append"](path, rule)
    return results


def _tmp_access_rule(
    method,
    ip=None,
    ttl=None,
    port=None,
    direction="in",
    port_origin="d",
    ip_origin="d",
    comment="",
):
    """
    Handles the cmd execution for tempdeny and tempallow commands.
    """
    if _status_csf():
        if ip is None:
            return {"error": "You must supply an ip address or CIDR."}
        if ttl is None:
            return {"error": "You must supply a ttl."}
        args = _build_tmp_access_args(method, ip, ttl, port, direction, comment)
        return __csf_cmd(args)


def _build_tmp_access_args(method, ip, ttl, port, direction, comment):
    """
    Builds the cmd args for temporary access/deny opts.
    """
    opt = _get_opt(method)
    args = "{} {} {}".format(opt, ip, ttl)
    if port:
        args += " -p {}".format(port)
    if direction:
        args += " -d {}".format(direction)
    if comment:
        args += " #{}".format(comment)
    return args


def running():
    """
    Check csf status

    CLI Example:

    .. code-block:: bash

        salt '*' csf.running
    """
    return _status_csf()


def disable():
    """
    Disable csf permanently

    CLI Example:

    .. code-block:: bash

        salt '*' csf.disable
    """
    if _status_csf():
        return __csf_cmd("-x")


def enable():
    """
    Activate csf if not running

    CLI Example:

    .. code-block:: bash

        salt '*' csf.enable
    """
    if not _status_csf():
        return __csf_cmd("-e")


def reload():
    """
    Restart csf

    CLI Example:

    .. code-block:: bash

        salt '*' csf.reload
    """
    return __csf_cmd("-r")


def tempallow(ip=None, ttl=None, port=None, direction=None, comment=""):
    """
    Add an rule to the temporary ip allow list.
    See :func:`_access_rule`.
    1- Add an IP:

    CLI Example:

    .. code-block:: bash

        salt '*' csf.tempallow 127.0.0.1 3600 port=22 direction='in' comment='# Temp dev ssh access'
    """
    return _tmp_access_rule("tempallow", ip, ttl, port, direction, comment)


def tempdeny(ip=None, ttl=None, port=None, direction=None, comment=""):
    """
    Add a rule to the temporary ip deny list.
    See :func:`_access_rule`.
    1- Add an IP:

    CLI Example:

    .. code-block:: bash

        salt '*' csf.tempdeny 127.0.0.1 300 port=22 direction='in' comment='# Brute force attempt'
    """
    return _tmp_access_rule("tempdeny", ip, ttl, port, direction, comment)


def allow(
    ip,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="s",
    ttl=None,
    comment="",
):
    """
    Add an rule to csf allowed hosts
    See :func:`_access_rule`.
    1- Add an IP:

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow 127.0.0.1
        salt '*' csf.allow 127.0.0.1 comment="Allow localhost"
    """
    return _access_rule(
        "allow",
        ip,
        port=port,
        proto=proto,
        direction=direction,
        port_origin=port_origin,
        ip_origin=ip_origin,
        comment=comment,
    )


def deny(
    ip,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="d",
    ttl=None,
    comment="",
):
    """
    Add an rule to csf denied hosts
    See :func:`_access_rule`.
    1- Deny an IP:

    CLI Example:

    .. code-block:: bash

        salt '*' csf.deny 127.0.0.1
        salt '*' csf.deny 127.0.0.1 comment="Too localhosty"
    """
    return _access_rule(
        "deny", ip, port, proto, direction, port_origin, ip_origin, comment
    )


def remove_temp_rule(ip):
    opt = _get_opt("temprm")
    args = "{} {}".format(opt, ip)
    return __csf_cmd(args)


def unallow(ip):
    """
    Remove a rule from the csf denied hosts
    See :func:`_access_rule`.
    1- Deny an IP:

    CLI Example:

    .. code-block:: bash

        salt '*' csf.unallow 127.0.0.1
    """
    return _access_rule("unallow", ip)


def undeny(ip):
    """
    Remove a rule from the csf denied hosts
    See :func:`_access_rule`.
    1- Deny an IP:

    CLI Example:

    .. code-block:: bash

        salt '*' csf.undeny 127.0.0.1
    """
    return _access_rule("undeny", ip)


def remove_rule(
    method,
    ip,
    port=None,
    proto="tcp",
    direction="in",
    port_origin="d",
    ip_origin="s",
    ttl=None,
    comment="",
):

    if method.startswith("temp") or ttl:
        return remove_temp_rule(ip)

    if not port:
        if method == "allow":
            return unallow(ip)
        elif method == "deny":
            return undeny(ip)

    if port:
        return _remove_access_rule_with_port(
            method=method,
            ip=ip,
            port=port,
            proto=proto,
            direction=direction,
            port_origin=port_origin,
            ip_origin=ip_origin,
        )


def allow_ports(ports, proto="tcp", direction="in"):
    """
    Fully replace the incoming or outgoing ports
    line in the csf.conf file - e.g. TCP_IN, TCP_OUT,
    UDP_IN, UDP_OUT, etc.

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow_ports ports="[22,80,443,4505,4506]" proto='tcp' direction='in'
    """

    results = []
    ports = set(ports)
    ports = list(ports)
    proto = proto.upper()
    direction = direction.upper()
    _validate_direction_and_proto(direction, proto)
    ports_csv = ",".join(map(str, ports))
    directions = build_directions(direction)

    for direction in directions:
        result = __salt__["file.replace"](
            "/etc/csf/csf.conf",
            # pylint: disable=anomalous-backslash-in-string
            pattern=r'^{}_{}(\ +)?\=(\ +)?".*"$'.format(proto, direction),
            # pylint: enable=anomalous-backslash-in-string
            repl='{}_{} = "{}"'.format(proto, direction, ports_csv),
        )
        results.append(result)

    return results


def get_ports(proto="tcp", direction="in"):
    """
    Lists ports from csf.conf based on direction and protocol.
    e.g. - TCP_IN, TCP_OUT, UDP_IN, UDP_OUT, etc..

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow_port 22 proto='tcp' direction='in'
    """

    proto = proto.upper()
    direction = direction.upper()
    results = {}
    _validate_direction_and_proto(direction, proto)
    directions = build_directions(direction)
    for direction in directions:
        option = "{}_{}".format(proto, direction)
        results[direction] = _csf_to_list(option)

    return results


def _validate_direction_and_proto(direction, proto):
    if direction.upper() not in ["IN", "OUT", "BOTH"]:
        raise SaltInvocationError("You must supply a direction of in, out, or both")
    if proto.upper() not in ["TCP", "UDP", "TCP6", "UDP6"]:
        raise SaltInvocationError(
            "You must supply tcp, udp, tcp6, or udp6 for the proto keyword"
        )
    return


def build_directions(direction):
    direction = direction.upper()
    if direction == "BOTH":
        directions = ["IN", "OUT"]
    else:
        directions = [direction]
    return directions


def allow_port(port, proto="tcp", direction="both"):
    """
    Like allow_ports, but it will append to the
    existing entry instead of replacing it.
    Takes a single port instead of a list of ports.

    CLI Example:

    .. code-block:: bash

        salt '*' csf.allow_port 22 proto='tcp' direction='in'
    """

    ports = get_ports(proto=proto, direction=direction)
    direction = direction.upper()
    _validate_direction_and_proto(direction, proto)
    directions = build_directions(direction)
    results = []
    for direction in directions:
        _ports = ports[direction]
        _ports.append(port)
        results += allow_ports(_ports, proto=proto, direction=direction)
    return results


def get_testing_status():
    testing = _csf_to_list("TESTING")[0]
    return testing


def _toggle_testing(val):
    if val == "on":
        val = "1"
    elif val == "off":
        val = "0"
    else:
        raise SaltInvocationError("Only valid arg is 'on' or 'off' here.")

    result = __salt__["file.replace"](
        "/etc/csf/csf.conf",
        pattern=r'^TESTING(\ +)?\=(\ +)?".*"',  # pylint: disable=W1401
        repl='TESTING = "{}"'.format(val),
    )
    return result


def enable_testing_mode():
    return _toggle_testing("on")


def disable_testing_mode():
    return _toggle_testing("off")



"""
    :codeauthor: :email:`Shane Lee <slee@saltstack.com>`
"""

import salt.grains.lvm as lvm
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LvmGrainsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for LVM grains
    """

    def setup_loader_modules(self):
        return {
            lvm: {"__salt__": {}},
        }

    def test__linux_lvm(self):
        """
        Test grains._linux_lvm, normal return
        Should return a populated dictionary
        """

        vgs_out = "  vg00\n  vg01"
        lvs_out_vg00 = "  root\n  swap\n  tmp \n  usr \n  var"
        lvs_out_vg01 = "  opt \n"
        cmd_out = MagicMock(
            autospec=True, side_effect=[vgs_out, lvs_out_vg00, lvs_out_vg01]
        )

        patch_which = patch(
            "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
        )
        patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run": cmd_out})
        with patch_which, patch_cmd_lvm:
            ret = lvm._linux_lvm()

        assert ret == {
            "lvm": {"vg00": ["root", "swap", "tmp", "usr", "var"], "vg01": ["opt"]}
        }, ret

    def test__linux_lvm_no_lvm(self):
        """
        Test grains._linux_lvm, no lvm installed
        Should return nothing
        """

        vgs_out = "  vg00\n  vg01"
        lvs_out_vg00 = "  root\n  swap\n  tmp \n  usr \n  var"
        lvs_out_vg01 = "  opt \n"
        cmd_out = MagicMock(
            autospec=True, side_effect=[vgs_out, lvs_out_vg00, lvs_out_vg01]
        )

        patch_which = patch("salt.utils.path.which", autospec=True, return_value="")
        patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run": cmd_out})
        with patch_which, patch_cmd_lvm:
            ret = lvm._linux_lvm()

        assert ret is None, ret

    def test__linux_lvm_no_logical_volumes(self):
        """
        Test grains._linux_lvm, lvm is installed but no volumes
        Should return a dictionary only with the header
        """

        vgs_out = ""
        cmd_out = MagicMock(autospec=True, side_effect=[vgs_out])

        patch_which = patch(
            "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lvm"
        )
        patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run": cmd_out})
        with patch_which, patch_cmd_lvm:
            ret = lvm._linux_lvm()

        assert ret == {"lvm": {}}, ret

    def test__aix_lvm(self):
        """
        Test grains._aix_lvm, normal return
        Should return a populated dictionary
        """

        lsvg_out = "rootvg\nothervg"
        lsvg_out_rootvg = (
            "rootvg:\nLV NAME             TYPE       LPs     PPs     PVs  LV STATE     "
            " MOUNT POINT\nhd5                 boot       1       1       1   "
            " closed/syncd  N/A\nhd6                 paging     32      32      1   "
            " open/syncd    N/A\nhd8                 jfs2log    1       1       1   "
            " open/syncd    N/A\nhd4                 jfs2       32      32      1   "
            " open/syncd    /\nhd2                 jfs2       16      16      1   "
            " open/syncd    /usr\nhd9var              jfs2       32      32      1   "
            " open/syncd    /var\nhd3                 jfs2       32      32      1   "
            " open/syncd    /tmp\nhd1                 jfs2       16      16      1   "
            " open/syncd    /home\nhd10opt             jfs2       16      16      1   "
            " open/syncd    /opt"
        )
        lsvg_out_othervg = (
            "othervg:\nLV NAME             TYPE       LPs     PPs     PVs  LV STATE    "
            "  MOUNT POINT\nloglv01             jfs2log    1       1       1   "
            " open/syncd    N/A\ndatalv              jfs2       16      16      1   "
            " open/syncd    /data"
        )
        cmd_out = MagicMock(
            autospec=True, side_effect=[lsvg_out, lsvg_out_rootvg, lsvg_out_othervg]
        )

        patch_which = patch(
            "salt.utils.path.which", autospec=True, return_value="/usr/sbin/lsvg"
        )
        patch_cmd_lvm = patch.dict(lvm.__salt__, {"cmd.run": cmd_out})
        with patch_which, patch_cmd_lvm:
            ret = lvm._aix_lvm()

        assert ret == {
            "lvm": {
                "rootvg": [
                    "hd5",
                    "hd6",
                    "hd8",
                    "hd4",
                    "hd2",
                    "hd9var",
                    "hd3",
                    "hd1",
                    "hd10opt",
                ],
                "othervg": ["loglv01", "datalv"],
            }
        }, ret



"""
    :codeauthor: Anthony Shaw <anthonyshaw@apache.org>
"""


import salt.modules.napalm_network as napalm_network
import tests.support.napalm as napalm_test_support
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class NapalmNetworkModuleTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        patcher = patch(
            "salt.utils.napalm.get_device",
            MagicMock(return_value=napalm_test_support.MockNapalmDevice()),
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        module_globals = {
            "__salt__": {
                "config.get": MagicMock(
                    return_value={"test": {"driver": "test", "key": "2orgk34kgk34g"}}
                ),
                "file.file_exists": napalm_test_support.true,
                "file.join": napalm_test_support.join,
                "file.get_managed": napalm_test_support.get_managed_file,
                "random.hash": napalm_test_support.random_hash,
            }
        }

        return {napalm_network: module_globals}

    def test_connected_pass(self):
        ret = napalm_network.connected()
        assert ret["out"] is True

    def test_facts(self):
        ret = napalm_network.facts()
        assert ret["out"] == napalm_test_support.TEST_FACTS.copy()

    def test_environment(self):
        ret = napalm_network.environment()
        assert ret["out"] == napalm_test_support.TEST_ENVIRONMENT.copy()

    def test_cli_single_command(self):
        """
        Test that CLI works with 1 arg
        """
        ret = napalm_network.cli("show run")
        assert ret["out"] == napalm_test_support.TEST_COMMAND_RESPONSE.copy()

    def test_cli_multi_command(self):
        """
        Test that CLI works with 2 arg
        """
        ret = napalm_network.cli("show run", "show run")
        assert ret["out"] == napalm_test_support.TEST_COMMAND_RESPONSE.copy()

    def test_traceroute(self):
        ret = napalm_network.traceroute("destination.com")
        assert list(ret["out"].keys())[0] == "success"

    def test_ping(self):
        ret = napalm_network.ping("destination.com")
        assert list(ret["out"].keys())[0] == "success"

    def test_arp(self):
        ret = napalm_network.arp()
        assert ret["out"] == napalm_test_support.TEST_ARP_TABLE.copy()

    def test_ipaddrs(self):
        ret = napalm_network.ipaddrs()
        assert ret["out"] == napalm_test_support.TEST_IPADDRS.copy()

    def test_interfaces(self):
        ret = napalm_network.interfaces()
        assert ret["out"] == napalm_test_support.TEST_INTERFACES.copy()

    def test_lldp(self):
        ret = napalm_network.lldp()
        assert ret["out"] == napalm_test_support.TEST_LLDP_NEIGHBORS.copy()

    def test_mac(self):
        ret = napalm_network.mac()
        assert ret["out"] == napalm_test_support.TEST_MAC_TABLE.copy()

    def test_config(self):
        ret = napalm_network.config("running")
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()

    def test_optics(self):
        ret = napalm_network.optics()
        assert ret["out"] == napalm_test_support.TEST_OPTICS.copy()

    def test_load_config(self):
        ret = napalm_network.load_config(text="new config")
        assert ret["result"]

    def test_load_config_replace(self):
        ret = napalm_network.load_config(text="new config", replace=True)
        assert ret["result"]

    def test_load_template(self):
        ret = napalm_network.load_template("set_ntp_peers", peers=["192.168.0.1"])
        assert ret["out"] is None

    def test_commit(self):
        ret = napalm_network.commit()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()

    def test_discard_config(self):
        ret = napalm_network.discard_config()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()

    def test_compare_config(self):
        ret = napalm_network.compare_config()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()

    def test_rollback(self):
        ret = napalm_network.rollback()
        assert ret["out"] == napalm_test_support.TEST_RUNNING_CONFIG.copy()

    def test_config_changed(self):
        ret = napalm_network.config_changed()
        assert ret == (True, "")

    def test_config_control(self):
        ret = napalm_network.config_control()
        assert ret == (True, "")



"""
Control virtual machines via Salt
"""


import logging
import os.path

import salt.client
import salt.key
import salt.utils.cloud
import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import SaltClientError

log = logging.getLogger(__name__)


def _determine_host(data, omit=""):
    """
    Determine what the most resource free host is based on the given data
    """
    # This is just checking for the host with the most free ram, this needs
    # to be much more complicated.
    host = ""
    bestmem = 0
    for hv_, comps in data.items():
        if hv_ == omit:
            continue
        if not isinstance(comps, dict):
            continue
        if comps.get("freemem", 0) > bestmem:
            bestmem = comps["freemem"]
            host = hv_
    return host


def _find_vm(name, data, quiet=False):
    """
    Scan the query data for the named VM
    """
    for hv_ in data:
        # Check if data is a dict, and not '"virt.full_info" is not available.'
        if not isinstance(data[hv_], dict):
            continue
        if name in data[hv_].get("vm_info", {}):
            ret = {hv_: {name: data[hv_]["vm_info"][name]}}
            if not quiet:
                __jid_event__.fire_event(
                    {"data": ret, "outputter": "nested"}, "progress"
                )
            return ret
    return {}


def query(host=None, quiet=False):
    """
    Query the virtual machines. When called without options all hosts
    are detected and a full query is returned. A single host can be
    passed in to specify an individual host to query.
    """
    if quiet:
        log.warning("'quiet' is deprecated. Please migrate to --quiet")
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            for info in client.cmd_iter(
                "virtual:physical", "virt.full_info", tgt_type="grain"
            ):
                if not info:
                    continue
                if not isinstance(info, dict):
                    continue
                chunk = {}
                id_ = next(iter(info.keys()))
                if host:
                    if host != id_:
                        continue
                if not isinstance(info[id_], dict):
                    continue
                if "ret" not in info[id_]:
                    continue
                if not isinstance(info[id_]["ret"], dict):
                    continue
                chunk[id_] = info[id_]["ret"]
                ret.update(chunk)
                if not quiet:
                    __jid_event__.fire_event(
                        {"data": chunk, "outputter": "virt_query"}, "progress"
                    )
        except SaltClientError as client_error:
            print(client_error)
        return ret


def list(host=None, quiet=False, hyper=None):  # pylint: disable=redefined-builtin
    """
    List the virtual machines on each host, this is a simplified query,
    showing only the virtual machine names belonging to each host.
    A single host can be passed in to specify an individual host
    to list.
    """
    if quiet:
        log.warning("'quiet' is deprecated. Please migrate to --quiet")
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        for info in client.cmd_iter(
            "virtual:physical", "virt.vm_info", tgt_type="grain"
        ):
            if not info:
                continue
            if not isinstance(info, dict):
                continue
            chunk = {}
            id_ = next(iter(info.keys()))
            if host:
                if host != id_:
                    continue
            if not isinstance(info[id_], dict):
                continue
            if "ret" not in info[id_]:
                continue
            if not isinstance(info[id_]["ret"], dict):
                continue
            data = {}
            for key, val in info[id_]["ret"].items():
                if val["state"] in data:
                    data[val["state"]].append(key)
                else:
                    data[val["state"]] = [key]
            chunk[id_] = data
            ret.update(chunk)
            if not quiet:
                __jid_event__.fire_event(
                    {"data": chunk, "outputter": "nested"}, "progress"
                )

        return ret


def next_host():
    """
    Return the host to use for the next autodeployed VM. This queries
    the available host and executes some math the determine the most
    "available" next host.
    """
    host = _determine_host(query(quiet=True))
    print(host)
    return host


def host_info(host=None):
    """
    Return information about the host connected to this master
    """
    data = query(host, quiet=True)
    for id_ in data:
        if "vm_info" in data[id_]:
            data[id_].pop("vm_info")
    __jid_event__.fire_event({"data": data, "outputter": "nested"}, "progress")
    return data


def init(
    name,
    cpu,
    mem,
    image,
    hypervisor="kvm",
    host=None,
    seed=True,
    nic="default",
    install=True,
    start=True,
    disk="default",
    saltenv="base",
    enable_vnc=False,
    seed_cmd="seed.apply",
    enable_qcow=False,
    serial_type="None",
):
    """
    This routine is used to create a new virtual machine. This routines takes
    a number of options to determine what the newly created virtual machine
    will look like.

    name
        The mandatory name of the new virtual machine. The name option is
        also the minion id, all minions must have an id.

    cpu
        The number of cpus to allocate to this new virtual machine.

    mem
        The amount of memory to allocate to this virtual machine. The number
        is interpreted in megabytes.

    image
        The network location of the virtual machine image, commonly a location
        on the salt fileserver, but http, https and ftp can also be used.

    hypervisor
        The hypervisor to use for the new virtual machine. Default is `kvm`.

    host
        The host to use for the new virtual machine, if this is omitted
        Salt will automatically detect what host to use.

    seed
        Set to `False` to prevent Salt from seeding the new virtual machine.

    nic
        The nic profile to use, defaults to the "default" nic profile which
        assumes a single network interface per VM associated with the "br0"
        bridge on the master.

    install
        Set to False to prevent Salt from installing a minion on the new VM
        before it spins up.

    disk
        The disk profile to use

    saltenv
        The Salt environment to use

    enable_vnc
        Whether a VNC screen is attached to resulting VM. Default is `False`.

    seed_cmd
        If seed is `True`, use this execution module function to seed new VM.
        Default is `seed.apply`.

    enable_qcow
        Clone disk image as a copy-on-write qcow2 image, using downloaded
        `image` as backing file.

    serial_type
        Enable serial console. Set to 'pty' for serial console or 'tcp' for
        telnet.
        Default is 'None'
    """
    __jid_event__.fire_event({"message": "Searching for hosts"}, "progress")
    data = query(host, quiet=True)
    # Check if the name is already deployed
    for node in data:
        if "vm_info" in data[node]:
            if name in data[node]["vm_info"]:
                __jid_event__.fire_event(
                    {"message": "Virtual machine {} is already deployed".format(name)},
                    "progress",
                )
                return "fail"

    if host is None:
        host = _determine_host(data)

    if host not in data or not host:
        __jid_event__.fire_event(
            {"message": "Host {} was not found".format(host)}, "progress"
        )
        return "fail"

    pub_key = None
    priv_key = None
    if seed:
        __jid_event__.fire_event({"message": "Minion will be preseeded"}, "progress")
        priv_key, pub_key = salt.utils.cloud.gen_keys()
        accepted_key = os.path.join(__opts__["pki_dir"], "minions", name)
        with salt.utils.files.fopen(accepted_key, "w") as fp_:
            fp_.write(salt.utils.stringutils.to_str(pub_key))

    with salt.client.get_local_client(__opts__["conf_file"]) as client:

        __jid_event__.fire_event(
            {"message": "Creating VM {} on host {}".format(name, host)}, "progress"
        )
        try:
            cmd_ret = client.cmd_iter(
                host,
                "virt.init",
                [name, cpu, mem],
                timeout=600,
                kwarg={
                    "image": image,
                    "nic": nic,
                    "hypervisor": hypervisor,
                    "start": start,
                    "disk": disk,
                    "saltenv": saltenv,
                    "seed": seed,
                    "install": install,
                    "pub_key": pub_key,
                    "priv_key": priv_key,
                    "seed_cmd": seed_cmd,
                    "enable_vnc": enable_vnc,
                    "enable_qcow": enable_qcow,
                    "serial_type": serial_type,
                },
            )
        except SaltClientError as client_error:
            # Fall through to ret error handling below
            print(client_error)

        ret = next(cmd_ret)
        if not ret:
            __jid_event__.fire_event(
                {"message": "VM {} was not initialized.".format(name)}, "progress"
            )
            return "fail"
        for minion_id in ret:
            if ret[minion_id]["ret"] is False:
                print(
                    "VM {} initialization failed. Returned error: {}".format(
                        name, ret[minion_id]["ret"]
                    )
                )
                return "fail"

        __jid_event__.fire_event(
            {"message": "VM {} initialized on host {}".format(name, host)}, "progress"
        )
        return "good"


def vm_info(name, quiet=False):
    """
    Return the information on the named VM
    """
    data = query(quiet=True)
    return _find_vm(name, data, quiet)


def reset(name):
    """
    Force power down and restart an existing VM
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        data = vm_info(name, quiet=True)
        if not data:
            __jid_event__.fire_event(
                {"message": "Failed to find VM {} to reset".format(name)}, "progress"
            )
            return "fail"
        host = next(iter(data.keys()))
        try:
            cmd_ret = client.cmd_iter(host, "virt.reset", [name], timeout=600)
            for comp in cmd_ret:
                ret.update(comp)
            __jid_event__.fire_event(
                {"message": "Reset VM {}".format(name)}, "progress"
            )
        except SaltClientError as client_error:
            print(client_error)
        return ret


def start(name):
    """
    Start a named virtual machine
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        data = vm_info(name, quiet=True)
        if not data:
            __jid_event__.fire_event(
                {"message": "Failed to find VM {} to start".format(name)}, "progress"
            )
            return "fail"
        host = next(iter(data.keys()))
        if data[host][name]["state"] == "running":
            print("VM {} is already running".format(name))
            return "bad state"
        try:
            cmd_ret = client.cmd_iter(host, "virt.start", [name], timeout=600)
        except SaltClientError as client_error:
            return "Virtual machine {} not started: {}".format(name, client_error)
        for comp in cmd_ret:
            ret.update(comp)
        __jid_event__.fire_event({"message": "Started VM {}".format(name)}, "progress")
        return "good"


def force_off(name):
    """
    Force power down the named virtual machine
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        data = vm_info(name, quiet=True)
        if not data:
            print("Failed to find VM {} to destroy".format(name))
            return "fail"
        host = next(iter(data.keys()))
        if data[host][name]["state"] == "shutdown":
            print("VM {} is already shutdown".format(name))
            return "bad state"
        try:
            cmd_ret = client.cmd_iter(host, "virt.stop", [name], timeout=600)
        except SaltClientError as client_error:
            return "Virtual machine {} could not be forced off: {}".format(
                name, client_error
            )
        for comp in cmd_ret:
            ret.update(comp)
        __jid_event__.fire_event(
            {"message": "Powered off VM {}".format(name)}, "progress"
        )
        return "good"


def purge(name, delete_key=True):
    """
    Destroy the named VM
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        data = vm_info(name, quiet=True)
        if not data:
            __jid_event__.fire_event(
                {"error": "Failed to find VM {} to purge".format(name)}, "progress"
            )
            return "fail"
        host = next(iter(data.keys()))
        try:
            cmd_ret = client.cmd_iter(host, "virt.purge", [name, True], timeout=600)
        except SaltClientError as client_error:
            return "Virtual machine {} could not be purged: {}".format(
                name, client_error
            )

        for comp in cmd_ret:
            ret.update(comp)

        if delete_key:
            log.debug("Deleting key %s", name)
            with salt.key.Key(__opts__) as skey:
                skey.delete_key(name)
        __jid_event__.fire_event({"message": "Purged VM {}".format(name)}, "progress")
        return "good"


def pause(name):
    """
    Pause the named VM
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:

        data = vm_info(name, quiet=True)
        if not data:
            __jid_event__.fire_event(
                {"error": "Failed to find VM {} to pause".format(name)}, "progress"
            )
            return "fail"
        host = next(iter(data.keys()))
        if data[host][name]["state"] == "paused":
            __jid_event__.fire_event(
                {"error": "VM {} is already paused".format(name)}, "progress"
            )
            return "bad state"
        try:
            cmd_ret = client.cmd_iter(host, "virt.pause", [name], timeout=600)
        except SaltClientError as client_error:
            return "Virtual machine {} could not be pasued: {}".format(
                name, client_error
            )
        for comp in cmd_ret:
            ret.update(comp)
        __jid_event__.fire_event({"message": "Paused VM {}".format(name)}, "progress")
        return "good"


def resume(name):
    """
    Resume a paused VM
    """
    ret = {}
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        data = vm_info(name, quiet=True)
        if not data:
            __jid_event__.fire_event(
                {"error": "Failed to find VM {} to pause".format(name)}, "progress"
            )
            return "not found"
        host = next(iter(data.keys()))
        if data[host][name]["state"] != "paused":
            __jid_event__.fire_event(
                {"error": "VM {} is not paused".format(name)}, "progress"
            )
            return "bad state"
        try:
            cmd_ret = client.cmd_iter(host, "virt.resume", [name], timeout=600)
        except SaltClientError as client_error:
            return "Virtual machine {} could not be resumed: {}".format(
                name, client_error
            )
        for comp in cmd_ret:
            ret.update(comp)
        __jid_event__.fire_event({"message": "Resumed VM {}".format(name)}, "progress")
        return "good"


def migrate(name, target=""):
    """
    Migrate a VM from one host to another. This routine will just start
    the migration and display information on how to look up the progress.
    """
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        data = query(quiet=True)
        origin_data = _find_vm(name, data, quiet=True)
        try:
            origin_host = next(iter(origin_data))
        except StopIteration:
            __jid_event__.fire_event(
                {"error": "Named VM {} was not found to migrate".format(name)},
                "progress",
            )
            return ""
        disks = origin_data[origin_host][name]["disks"]
        if not origin_data:
            __jid_event__.fire_event(
                {"error": "Named VM {} was not found to migrate".format(name)},
                "progress",
            )
            return ""
        if not target:
            target = _determine_host(data, origin_host)
        if target not in data:
            __jid_event__.fire_event(
                {"error": "Target host {} not found".format(origin_data)}, "progress"
            )
            return ""
        try:
            client.cmd(target, "virt.seed_non_shared_migrate", [disks, True])
            jid = client.cmd_async(
                origin_host, "virt.migrate", [name, target], copy_storage=all
            )
        except SaltClientError as client_error:
            return "Virtual machine {} could not be migrated: {}".format(
                name, client_error
            )

        msg = (
            "The migration of virtual machine {} to host {} has begun, "
            "and can be tracked via jid {}. The ``salt-run virt.query`` "
            "runner can also be used, the target VM will be shown as paused "
            "until the migration is complete.".format(name, target, jid)
        )
        __jid_event__.fire_event({"message": msg}, "progress")



#!/usr/bin/env python
"""
Pepa
====

Configuration templating for SaltStack using Hierarchical substitution and Jinja.

Configuring Pepa
================

.. code-block:: yaml

    extension_modules: /srv/salt/ext

    ext_pillar:
      - pepa:
          resource: host                # Name of resource directory and sub-key in pillars
          sequence:                     # Sequence used for hierarchical substitution
            - hostname:                 # Name of key
                name: input             # Alias used for template directory
                base_only: True         # Only use templates from Base environment, i.e. no staging
            - default:
            - environment:
            - location..region:
                name: region
            - location..country:
                name: country
            - location..datacenter:
                name: datacenter
            - roles:
            - osfinger:
                name: os
            - hostname:
                name: override
                base_only: True
          subkey: True                  # Create a sub-key in pillars, named after the resource in this case [host]
          subkey_only: True             # Only create a sub-key, and leave the top level untouched

    pepa_roots:                         # Base directory for each environment
      base: /srv/pepa/base              # Path for base environment
      dev: /srv/pepa/base               # Associate dev with base
      qa: /srv/pepa/qa
      prod: /srv/pepa/prod

    # Use a different delimiter for nested dictionaries, defaults to '..' since some keys may use '.' in the name
    #pepa_delimiter: ..

    # Supply Grains for Pepa, this should **ONLY** be used for testing or validation
    #pepa_grains:
    #  environment: dev

    # Supply Pillar for Pepa, this should **ONLY** be used for testing or validation
    #pepa_pillars:
    #  saltversion: 0.17.4

    # Enable debug for Pepa, and keep Salt on warning
    #log_level: debug

    #log_granular_levels:
    #  salt: warning
    #  salt.loaded.ext.pillar.pepa: debug

Pepa can also be used in Master-less SaltStack setup.

Command line
============

.. code-block:: bash

    usage: pepa.py [-h] [-c CONFIG] [-d] [-g GRAINS] [-p PILLAR] [-n] [-v]
                   hostname

    positional arguments:
      hostname              Hostname

    optional arguments:
      -h, --help            show this help message and exit
      -c CONFIG, --config CONFIG
                            Configuration file
      -d, --debug           Print debug info
      -g GRAINS, --grains GRAINS
                            Input Grains as YAML
      -p PILLAR, --pillar PILLAR
                            Input Pillar as YAML
      -n, --no-color        No color output
      -v, --validate        Validate output

Templates
=========

Templates is configuration for a host or software, that can use information from Grains or Pillars. These can then be used for hierarchically substitution.

**Example File:** host/input/test_example_com.yaml

.. code-block:: yaml

    location..region: emea
    location..country: nl
    location..datacenter: foobar
    environment: dev
    roles:
      - salt.master
    network..gateway: 10.0.0.254
    network..interfaces..eth0..hwaddr: 00:20:26:a1:12:12
    network..interfaces..eth0..dhcp: False
    network..interfaces..eth0..ipv4: 10.0.0.3
    network..interfaces..eth0..netmask: 255.255.255.0
    network..interfaces..eth0..fqdn: {{ hostname }}
    cobbler..profile: fedora-19-x86_64

As you see in this example you can use Jinja directly inside the template.

**Example File:** host/region/amer.yaml

.. code-block:: yaml

    network..dns..servers:
      - 10.0.0.1
      - 10.0.0.2
    time..ntp..servers:
      - ntp1.amer.example.com
      - ntp2.amer.example.com
      - ntp3.amer.example.com
    time..timezone: America/Chihuahua
    yum..mirror: yum.amer.example.com

Each template is named after the value of the key using lowercase and all extended characters are replaced with underscore.

**Example:**

osfinger: Fedora-19

**Would become:**

fedora_19.yaml

Nested dictionaries
===================

In order to create nested dictionaries as output you can use double dot **".."** as a delimiter. You can change this using "pepa_delimiter" we choose double dot since single dot is already used by key names in some modules, and using ":" requires quoting in the YAML.

**Example:**

.. code-block:: yaml

    network..dns..servers:
      - 10.0.0.1
      - 10.0.0.2
    network..dns..options:
      - timeout:2
      - attempts:1
      - ndots:1
    network..dns..search:
      - example.com

**Would become:**

.. code-block:: yaml

    network:
      dns:
        servers:
          - 10.0.0.1
          - 10.0.0.2
        options:
          - timeout:2
          - attempts:1
          - ndots:1
        search:
          - example.com

Operators
=========

Operators can be used to merge/unset a list/hash or set the key as immutable, so it can't be changed.

=========== ================================================
Operator    Description
=========== ================================================
merge()     Merge list or hash
unset()     Unset key
immutable() Set the key as immutable, so it can't be changed
imerge()    Set immutable and merge
iunset()    Set immutable and unset
=========== ================================================

**Example:**

.. code-block:: yaml

    network..dns..search..merge():
      - foobar.com
      - dummy.nl
    owner..immutable(): Operations
    host..printers..unset():

Validation
==========

Since it's very hard to test Jinja as is, the best approach is to run all the permutations of input and validate the output, i.e. Unit Testing.

To facilitate this in Pepa we use YAML, Jinja and Cerberus <https://github.com/nicolaiarocci/cerberus>.

Schema
======

So this is a validation schema for network configuration, as you see it can be customized with Jinja just as Pepa templates.

This was designed to be run as a build job in Jenkins or similar tool. You can provide Grains/Pillar input using either the config file or command line arguments.

**File Example: host/validation/network.yaml**

.. code-block:: jinja

    network..dns..search:
      type: list
      allowed:
        - example.com

    network..dns..options:
      type: list
      allowed: ['timeout:2', 'attempts:1', 'ndots:1']

    network..dns..servers:
      type: list
      schema:
        regex: ^([0-9]{1,3}\\.){3}[0-9]{1,3}$

    network..gateway:
      type: string
      regex: ^([0-9]{1,3}\\.){3}[0-9]{1,3}$

    {% if network.interfaces is defined %}
    {% for interface in network.interfaces %}

    network..interfaces..{{ interface }}..dhcp:
      type: boolean

    network..interfaces..{{ interface }}..fqdn:
      type: string
      regex: ^([a-z0-9]([a-z0-9-]{0,61}[a-z0-9])?\\.)+[a-zA-Z]{2,6}$

    network..interfaces..{{ interface }}..hwaddr:
      type: string
      regex: ^([0-9a-f]{1,2}\\:){5}[0-9a-f]{1,2}$

    network..interfaces..{{ interface }}..ipv4:
      type: string
      regex: ^([0-9]{1,3}\\.){3}[0-9]{1,3}$

    network..interfaces..{{ interface }}..netmask:
      type: string
      regex: ^([0-9]{1,3}\\.){3}[0-9]{1,3}$

    {% endfor %}
    {% endif %}

Links
=====

For more examples and information see <https://github.com/mickep76/pepa>.
"""

# Import futures

import glob
import logging
import os
import re
import sys

import jinja2
import salt.utils.files
import salt.utils.yaml

__author__ = "Michael Persson <michael.ake.persson@gmail.com>"
__copyright__ = "Copyright (c) 2013 Michael Persson"
__license__ = "Apache License, Version 2.0"
__version__ = "0.6.6"


try:
    import requests

    HAS_REQUESTS = True
except ImportError:
    HAS_REQUESTS = False


# Only used when called from a terminal
log = None
if __name__ == "__main__":
    import argparse  # pylint: disable=minimum-python-version

    parser = argparse.ArgumentParser()
    parser.add_argument("hostname", help="Hostname")
    parser.add_argument(
        "-c", "--config", default="/etc/salt/master", help="Configuration file"
    )
    parser.add_argument("-d", "--debug", action="store_true", help="Print debug info")
    parser.add_argument("-g", "--grains", help="Input Grains as YAML")
    parser.add_argument("-p", "--pillar", help="Input Pillar as YAML")
    parser.add_argument("-n", "--no-color", action="store_true", help="No color output")
    parser.add_argument("-v", "--validate", action="store_true", help="Validate output")
    parser.add_argument(
        "-q",
        "--query-api",
        action="store_true",
        help="Query Saltstack REST API for Grains",
    )
    parser.add_argument(
        "--url", default="https://salt:8000", help="URL for SaltStack REST API"
    )
    parser.add_argument("-u", "--username", help="Username for SaltStack REST API")
    parser.add_argument("-P", "--password", help="Password for SaltStack REST API")
    args = parser.parse_args()

    LOG_LEVEL = logging.WARNING
    if args.debug:
        LOG_LEVEL = logging.DEBUG

    formatter = None
    if not args.no_color:
        try:
            import colorlog  # pylint: disable=import-error

            formatter = colorlog.ColoredFormatter(
                "[%(log_color)s%(levelname)-8s%(reset)s]"
                " %(log_color)s%(message)s%(reset)s"
            )
        except ImportError:
            formatter = logging.Formatter("[%(levelname)-8s] %(message)s")
    else:
        formatter = logging.Formatter("[%(levelname)-8s] %(message)s")

    stream = logging.StreamHandler()
    stream.setLevel(LOG_LEVEL)
    stream.setFormatter(formatter)

    log = logging.getLogger("pythonConfig")
    log.setLevel(LOG_LEVEL)
    log.addHandler(stream)
else:
    log = logging.getLogger(__name__)


# Options
__opts__ = {
    "pepa_roots": {"base": "/srv/salt"},
    "pepa_delimiter": "..",
    "pepa_validate": False,
}


def __virtual__():
    """
    Only return if all the modules are available
    """
    if not HAS_REQUESTS:
        return False

    return True


def key_value_to_tree(data):
    """
    Convert key/value to tree
    """
    tree = {}
    for flatkey, value in data.items():
        t = tree
        keys = flatkey.split(__opts__["pepa_delimiter"])
        for i, key in enumerate(keys, 1):
            if i == len(keys):
                t[key] = value
            else:
                t = t.setdefault(key, {})
    return tree


def ext_pillar(minion_id, pillar, resource, sequence, subkey=False, subkey_only=False):
    """
    Evaluate Pepa templates
    """
    roots = __opts__["pepa_roots"]

    # Default input
    inp = {}
    inp["default"] = "default"
    inp["hostname"] = minion_id

    if "environment" in pillar:
        inp["environment"] = pillar["environment"]
    elif "environment" in __grains__:
        inp["environment"] = __grains__["environment"]
    else:
        inp["environment"] = "base"

    # Load templates
    output = inp
    output["pepa_templates"] = []
    immutable = {}

    for categ, info in [next(iter(s.items())) for s in sequence]:
        if categ not in inp:
            log.warning("Category is not defined: %s", categ)
            continue

        alias = None
        if isinstance(info, dict) and "name" in info:
            alias = info["name"]
        else:
            alias = categ

        templdir = None
        if info and "base_only" in info and info["base_only"]:
            templdir = os.path.join(roots["base"], resource, alias)
        else:
            templdir = os.path.join(roots[inp["environment"]], resource, alias)

        entries = []
        if isinstance(inp[categ], list):
            entries = inp[categ]
        elif not inp[categ]:
            log.warning("Category has no value set: %s", categ)
            continue
        else:
            entries = [inp[categ]]

        for entry in entries:
            results_jinja = None
            results = None
            fn = os.path.join(templdir, re.sub(r"\W", "_", entry.lower()) + ".yaml")
            if os.path.isfile(fn):
                log.info("Loading template: %s", fn)
                with salt.utils.files.fopen(fn) as fhr:
                    template = jinja2.Template(fhr.read())
                output["pepa_templates"].append(fn)

                try:
                    data = key_value_to_tree(output)
                    data["grains"] = __grains__.copy()
                    data["pillar"] = pillar.copy()
                    results_jinja = template.render(data)
                    results = salt.utils.yaml.safe_load(results_jinja)
                except jinja2.UndefinedError as err:
                    log.error("Failed to parse JINJA template: %s\n%s", fn, err)
                except salt.utils.yaml.YAMLError as err:
                    log.error("Failed to parse YAML in template: %s\n%s", fn, err)
            else:
                log.info("Template doesn't exist: %s", fn)
                continue

            if results is not None:
                for key in results:
                    skey = key.rsplit(__opts__["pepa_delimiter"], 1)
                    rkey = None
                    operator = None
                    if len(skey) > 1 and key.rfind("()") > 0:
                        rkey = skey[0].rstrip(__opts__["pepa_delimiter"])
                        operator = skey[1]

                    if key in immutable:
                        log.warning("Key %s is immutable, changes are not allowed", key)
                    elif rkey in immutable:
                        log.warning(
                            "Key %s is immutable, changes are not allowed", rkey
                        )
                    elif operator == "merge()" or operator == "imerge()":
                        if operator == "merge()":
                            log.debug("Merge key %s: %s", rkey, results[key])
                        else:
                            log.debug(
                                "Set immutable and merge key %s: %s", rkey, results[key]
                            )
                            immutable[rkey] = True
                        if rkey not in output:
                            log.error("Cant't merge key %s doesn't exist", rkey)
                        elif not isinstance(results[key], type(output[rkey])):
                            log.error("Can't merge different types for key %s", rkey)
                        elif isinstance(results[key], dict):
                            output[rkey].update(results[key])
                        elif isinstance(results[key], list):
                            output[rkey].extend(results[key])
                        else:
                            log.error(
                                "Unsupported type need to be list or dict for key %s",
                                rkey,
                            )
                    elif operator == "unset()" or operator == "iunset()":
                        if operator == "unset()":
                            log.debug("Unset key %s", rkey)
                        else:
                            log.debug("Set immutable and unset key %s", rkey)
                            immutable[rkey] = True
                        if rkey in output:
                            del output[rkey]
                    elif operator == "immutable()":
                        log.debug(
                            "Set immutable and substitute key %s: %s",
                            rkey,
                            results[key],
                        )
                        immutable[rkey] = True
                        output[rkey] = results[key]
                    elif operator is not None:
                        log.error(
                            "Unsupported operator %s, skipping key %s", operator, rkey
                        )
                    else:
                        log.debug("Substitute key %s: %s", key, results[key])
                        output[key] = results[key]

    tree = key_value_to_tree(output)
    pillar_data = {}
    if subkey_only:
        pillar_data[resource] = tree.copy()
    elif subkey:
        pillar_data = tree
        pillar_data[resource] = tree.copy()
    else:
        pillar_data = tree
    if __opts__["pepa_validate"]:
        pillar_data["pepa_keys"] = output.copy()
    return pillar_data


def validate(output, resource):
    """
    Validate Pepa templates
    """
    try:
        import cerberus  # pylint: disable=import-error
    except ImportError:
        log.critical("You need module cerberus in order to use validation")
        return

    roots = __opts__["pepa_roots"]

    valdir = os.path.join(roots["base"], resource, "validate")

    all_schemas = {}
    pepa_schemas = []
    for fn in glob.glob(valdir + "/*.yaml"):
        log.info("Loading schema: %s", fn)
        with salt.utils.files.fopen(fn) as fhr:
            template = jinja2.Template(fhr.read())
        data = output
        data["grains"] = __grains__.copy()
        data["pillar"] = __pillar__.copy()
        schema = salt.utils.yaml.safe_load(template.render(data))
        all_schemas.update(schema)
        pepa_schemas.append(fn)

    val = cerberus.Validator()
    if not val.validate(output["pepa_keys"], all_schemas):
        for ekey, error in val.errors.items():
            log.warning("Validation failed for key %s: %s", ekey, error)

    output["pepa_schema_keys"] = all_schemas
    output["pepa_schemas"] = pepa_schemas


# Only used when called from a terminal
if __name__ == "__main__":
    # Load configuration file
    if not os.path.isfile(args.config):
        log.critical("Configuration file doesn't exist: %s", args.config)
        sys.exit(1)

    # Get configuration
    with salt.utils.files.fopen(args.config) as fh_:
        __opts__.update(salt.utils.yaml.safe_load(fh_))

    loc = 0
    for name in [next(iter(list(e.keys()))) for e in __opts__["ext_pillar"]]:
        if name == "pepa":
            break
        loc += 1

    # Get grains
    __grains__ = {}
    if "pepa_grains" in __opts__:
        __grains__ = __opts__["pepa_grains"]
    if args.grains:
        __grains__.update(salt.utils.yaml.safe_load(args.grains))

    # Get pillars
    __pillar__ = {}
    if "pepa_pillar" in __opts__:
        __pillar__ = __opts__["pepa_pillar"]
    if args.pillar:
        __pillar__.update(salt.utils.yaml.safe_load(args.pillar))

    # Validate or not
    if args.validate:
        __opts__["pepa_validate"] = True

    if args.query_api:
        import requests
        import getpass

        username = args.username
        password = args.password
        if username is None:
            username = input("Username: ")
        if password is None:
            password = getpass.getpass()

        log.info("Authenticate REST API")
        auth = {"username": username, "password": password, "eauth": "pam"}
        request = requests.post(args.url + "/login", auth)

        if not request.ok:
            raise RuntimeError(
                "Failed to authenticate to SaltStack REST API: {}".format(request.text)
            )

        response = request.json()
        token = response["return"][0]["token"]

        log.info("Request Grains from REST API")
        headers = {"X-Auth-Token": token, "Accept": "application/json"}
        request = requests.get(args.url + "/minions/" + args.hostname, headers=headers)

        result = request.json().get("return", [{}])[0]
        if args.hostname not in result:
            raise RuntimeError("Failed to get Grains from SaltStack REST API")

        __grains__ = result[args.hostname]

    # Print results
    ex_subkey = False
    ex_subkey_only = False
    if "subkey" in __opts__["ext_pillar"][loc]["pepa"]:
        ex_subkey = __opts__["ext_pillar"][loc]["pepa"]["subkey"]
    if "subkey_only" in __opts__["ext_pillar"][loc]["pepa"]:
        ex_subkey_only = __opts__["ext_pillar"][loc]["pepa"]["subkey_only"]

    result = ext_pillar(
        args.hostname,
        __pillar__,
        __opts__["ext_pillar"][loc]["pepa"]["resource"],
        __opts__["ext_pillar"][loc]["pepa"]["sequence"],
        ex_subkey,
        ex_subkey_only,
    )

    if __opts__["pepa_validate"]:
        validate(result, __opts__["ext_pillar"][loc]["pepa"]["resource"])

    try:
        orig_ignore = salt.utils.yaml.SafeOrderedDumper.ignore_aliases
        salt.utils.yaml.SafeOrderedDumper.ignore_aliases = lambda x, y: True

        def _print_result(result):
            print(salt.utils.yaml.safe_dump(result, indent=4, default_flow_style=False))

        if not args.no_color:
            try:
                # pylint: disable=import-error
                import pygments
                import pygments.lexers
                import pygments.formatters

                # pylint: disable=no-member
                print(
                    pygments.highlight(
                        salt.utils.yaml.safe_dump(result),
                        pygments.lexers.YamlLexer(),
                        pygments.formatters.TerminalFormatter(),
                    )
                )
                # pylint: enable=no-member, import-error
            except ImportError:
                _print_result(result)
        else:
            _print_result(result)
    finally:
        # Undo monkeypatching
        salt.utils.yaml.SafeOrderedDumper.ignore_aliases = orig_ignore



"""
Management of Gentoo Overlays using layman
==========================================

A state module to manage Gentoo package overlays via layman

.. code-block:: yaml

    sunrise:
        layman.present
"""


def __virtual__():
    """
    Only load if the layman module is available in __salt__
    """
    if "layman.add" in __salt__:
        return "layman"
    return (False, "layman module could not be loaded")


def present(name):
    """
    Verify that the overlay is present

    name
        The name of the overlay to add
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    # Overlay already present
    if name in __salt__["layman.list_local"]():
        ret["comment"] = "Overlay {} already present".format(name)
    elif __opts__["test"]:
        ret["comment"] = "Overlay {} is set to be added".format(name)
        ret["result"] = None
        return ret
    else:
        # Does the overlay exist?
        if name not in __salt__["layman.list_all"]():
            ret["comment"] = "Overlay {} not found".format(name)
            ret["result"] = False
        else:
            # Attempt to add the overlay
            changes = __salt__["layman.add"](name)

            # The overlay failed to add
            if len(changes) < 1:
                ret["comment"] = "Overlay {} failed to add".format(name)
                ret["result"] = False
            # Success
            else:
                ret["changes"]["added"] = changes
                ret["comment"] = "Overlay {} added.".format(name)

    return ret


def absent(name):
    """
    Verify that the overlay is absent

    name
        The name of the overlay to delete
    """
    ret = {"changes": {}, "comment": "", "name": name, "result": True}

    # Overlay is already absent
    if name not in __salt__["layman.list_local"]():
        ret["comment"] = "Overlay {} already absent".format(name)
    elif __opts__["test"]:
        ret["comment"] = "Overlay {} is set to be deleted".format(name)
        ret["result"] = None
        return ret
    else:
        # Attempt to delete the overlay
        changes = __salt__["layman.delete"](name)

        # The overlay failed to delete
        if len(changes) < 1:
            ret["comment"] = "Overlay {} failed to delete".format(name)
            ret["result"] = False
        # Success
        else:
            ret["changes"]["deleted"] = changes
            ret["comment"] = "Overlay {} deleted.".format(name)

    return ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest
import salt.states.openstack_config as openstack_config
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {openstack_config: {"__opts__": {"test": False}}}


def test_present():
    """
    Test to ensure a value is set in an OpenStack configuration file.
    """
    name = "salt"
    filename = "/tmp/salt"
    section = "A"
    value = "SALT"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_lst = MagicMock(side_effect=[value, CommandExecutionError, "A"])
    mock_t = MagicMock(return_value=True)
    with patch.dict(
        openstack_config.__salt__,
        {"openstack_config.get": mock_lst, "openstack_config.set": mock_t},
    ):
        comt = "The value is already set to the correct value"
        ret.update({"comment": comt, "result": True})
        assert openstack_config.present(name, filename, section, value) == ret

        pytest.raises(
            CommandExecutionError,
            openstack_config.present,
            name,
            filename,
            section,
            value,
        )

        comt = "The value has been updated"
        ret.update({"comment": comt, "changes": {"Value": "Updated"}})
        assert openstack_config.present(name, filename, section, value) == ret


def test_absent():
    """
    Test to ensure a value is not set in an OpenStack configuration file.
    """
    name = "salt"
    filename = "/tmp/salt"
    section = "A"

    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    mock_lst = MagicMock(
        side_effect=[
            CommandExecutionError("parameter not found:"),
            CommandExecutionError,
            "A",
        ]
    )
    mock_t = MagicMock(return_value=True)
    with patch.dict(
        openstack_config.__salt__,
        {"openstack_config.get": mock_lst, "openstack_config.delete": mock_t},
    ):
        comt = "The value is already absent"
        ret.update({"comment": comt, "result": True})
        assert openstack_config.absent(name, filename, section) == ret

        pytest.raises(
            CommandExecutionError, openstack_config.absent, name, filename, section
        )

        comt = "The value has been deleted"
        ret.update({"comment": comt, "changes": {"Value": "Deleted"}})
        assert openstack_config.absent(name, filename, section) == ret



"""
Test the directory roster.
"""


import logging

import pytest
import salt.config
import salt.loader
import salt.roster.dir as dir_


@pytest.fixture
def roster_domain():
    return "test.roster.domain"


@pytest.fixture
def expected(roster_domain):
    return {
        "basic": {
            "test1_us-east-2_test_basic": {
                "host": "127.0.0.2",
                "port": 22,
                "sudo": True,
                "user": "scoundrel",
            }
        },
        "domain": {
            "test1_us-east-2_test_domain": {
                "host": "test1_us-east-2_test_domain." + roster_domain,
                "port": 2222,
                "user": "george",
            }
        },
        "empty": {
            "test1_us-east-2_test_empty": {
                "host": "test1_us-east-2_test_empty." + roster_domain,
            }
        },
    }


@pytest.fixture
def create_roster_files(tmp_path):
    badfile_contents = """
    #!jinja|yaml
    host: 127.0.0.2
    port: 22
    THIS FILE IS NOT WELL FORMED YAML
    sudo: true
    user: scoundrel
    """

    basic_contents = """
    #!jinja|yaml
    host: 127.0.0.2
    port: 22
    sudo: true
    user: scoundrel
    """

    domain_contents = """
    #!jinja|yaml
    port: 2222
    user: george
    """

    empty_contents = """
    """

    with pytest.helpers.temp_file(
        "test1_us-east-2_test_badfile", badfile_contents, directory=tmp_path
    ), pytest.helpers.temp_file(
        "test1_us-east-2_test_basic", basic_contents, directory=tmp_path
    ), pytest.helpers.temp_file(
        "test1_us-east-2_test_domain", domain_contents, directory=tmp_path
    ), pytest.helpers.temp_file(
        "test1_us-east-2_test_empty", empty_contents, directory=tmp_path
    ):
        yield


@pytest.fixture
def configure_loader_modules(roster_domain, salt_master_factory, tmp_path):

    opts = salt_master_factory.config.copy()
    utils = salt.loader.utils(opts, whitelist=["json", "stringutils", "roster_matcher"])
    runner = salt.loader.runner(opts, utils=utils, whitelist=["salt"])
    return {
        dir_: {
            "__opts__": {
                "extension_modules": "",
                "optimization_order": [0, 1, 2],
                "renderer": "jinja|yaml",
                "renderer_blacklist": [],
                "renderer_whitelist": [],
                "roster_dir": str(tmp_path),
                "roster_domain": roster_domain,
            },
            "__runner__": runner,
            "__utils__": utils,
        }
    }


def _test_match(ret, expected):
    """
    assertDictEquals is too strict with OrderedDicts. The order isn't crucial
    for roster entries, so we test that they contain the expected members directly.
    """
    assert ret != {}, "Found no matches, expected {}".format(expected)
    for minion, data in ret.items():
        assert minion in expected, "Expected minion {} to match, but it did not".format(
            minion
        )
        assert (
            dict(data) == expected[minion]
        ), "Data for minion {} did not match expectations".format(minion)


def test_basic_glob(expected, create_roster_files):
    """Test that minion files in the directory roster match and render."""
    expected = expected["basic"]
    ret = dir_.targets("*_basic", saltenv="")
    _test_match(ret, expected)


def test_basic_re(expected, create_roster_files):
    """Test that minion files in the directory roster match and render."""
    expected = expected["basic"]
    ret = dir_.targets(".*basic$", "pcre", saltenv="")
    _test_match(ret, expected)


def test_basic_list(expected, create_roster_files):
    """Test that minion files in the directory roster match and render."""
    expected = expected["basic"]
    ret = dir_.targets(expected.keys(), "list", saltenv="")
    _test_match(ret, expected)


def test_roster_domain(expected, create_roster_files):
    """Test that when roster_domain is configured, it will provide a default hostname
    in the roster of {filename}.{roster_domain}, so that users can use the minion
    id as the local hostname without having to supply the fqdn everywhere."""
    expected = expected["domain"]
    ret = dir_.targets(expected.keys(), "list", saltenv="")
    _test_match(ret, expected)


def test_empty(expected, create_roster_files):
    """Test that an empty roster file matches its hostname"""
    expected = expected["empty"]
    ret = dir_.targets("*_empty", saltenv="")
    _test_match(ret, expected)


def test_nomatch(create_roster_files):
    """Test that no errors happen when no files match"""
    try:
        ret = dir_.targets("", saltenv="")
    except:  # pylint: disable=bare-except
        pytest.fail(
            "No files matched, which is OK, but we raised an exception and we should not have."
        )
    assert len(ret) == 0, "Expected empty target list to yield zero targets."


def test_badfile(create_roster_files):
    """Test error handling when we can't render a file"""
    ret = dir_.targets("*badfile", saltenv="")
    assert len(ret) == 0


def test_badfile_logging(caplog, create_roster_files):
    """Test error handling when we can't render a file"""
    with caplog.at_level(logging.WARNING, logger="salt.roster.dir"):
        dir_.targets("*badfile", saltenv="")
        assert "test1_us-east-2_test_badfile" in caplog.text



import functools
import os
import subprocess

import pytest
import salt.utils.user

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_on_windows,
]


@pytest.fixture(scope="module")
def account_1():
    with pytest.helpers.create_account(create_group=True) as _account:
        yield _account


@pytest.fixture(scope="module")
def account_2(account_1):
    with pytest.helpers.create_account(group_name=account_1.group.name) as _account:
        yield _account


def test_chugid(account_1, tmp_path):

    # Since we're changing accounts to touch the file, the parent directory must be user and group writable
    tmp_path.chmod(0o770)

    testfile = tmp_path / "testfile"

    # We should fail because the parent directory group owner is not the account running the test
    ret = subprocess.run(
        ["touch", str(testfile)],
        preexec_fn=functools.partial(
            salt.utils.user.chugid_and_umask,
            runas=account_1.username,
            umask=None,
            group=None,
        ),
        check=False,
    )
    assert ret.returncode != 0

    # However if we change the group ownership to one of the account's groups, it should succeed
    os.chown(str(tmp_path), 0, account_1.group.info.gid)

    ret = subprocess.run(
        ["touch", str(testfile)],
        preexec_fn=functools.partial(
            salt.utils.user.chugid_and_umask,
            runas=account_1.username,
            umask=None,
            group=None,
        ),
        check=False,
    )
    assert ret.returncode == 0
    assert testfile.exists()
    testfile_stat = testfile.stat()
    assert testfile_stat.st_uid == account_1.info.uid
    assert testfile_stat.st_gid == account_1.info.gid


def test_chugid_and_group(account_1, account_2, tmp_path):

    # Since we're changing accounts to touch the file, the parent directory must be world-writable
    tmp_path.chmod(0o770)

    testfile = tmp_path / "testfile"

    # We should fail because the parent directory group owner is not the account running the test
    ret = subprocess.run(
        ["touch", str(testfile)],
        preexec_fn=functools.partial(
            salt.utils.user.chugid_and_umask,
            runas=account_2.username,
            umask=None,
            group=account_1.group.name,
        ),
        check=False,
    )
    assert ret.returncode != 0

    # However if we change the group ownership to one of the account's groups, it should succeed
    os.chown(str(tmp_path), 0, account_1.group.info.gid)

    ret = subprocess.run(
        ["touch", str(testfile)],
        preexec_fn=functools.partial(
            salt.utils.user.chugid_and_umask,
            runas=account_2.username,
            umask=None,
            group=account_1.group.name,
        ),
        check=False,
    )
    assert ret.returncode == 0
    assert testfile.exists()
    testfile_stat = testfile.stat()
    assert testfile_stat.st_uid == account_2.info.uid
    assert testfile_stat.st_gid == account_1.group.info.gid



"""
    :codeauthor: Joe Julian <me@joejulian.name>
"""
# Import the future

import logging

import salt.modules.tls as tls
from salt.utils.versions import LooseVersion
from tests.support.helpers import with_tempdir
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase, skipIf

try:
    # We're not going to actually use OpenSSL, we just want to check that
    # it's installed.
    import OpenSSL  # pylint: disable=unused-import

    NO_PYOPENSSL = False
except Exception:  # pylint: disable=broad-except
    NO_PYOPENSSL = True


log = logging.getLogger(__name__)

_TLS_TEST_DATA = {
    "ca_cert": """-----BEGIN CERTIFICATE-----
MIIEejCCA2KgAwIBAgIRANW6IUG5rMez0vPi3cSmS3QwDQYJKoZIhvcNAQELBQAw
eTELMAkGA1UEBhMCVVMxDTALBgNVBAgMBFV0YWgxFzAVBgNVBAcMDlNhbHQgTGFr
ZSBDaXR5MRIwEAYDVQQKDAlTYWx0U3RhY2sxEjAQBgNVBAMMCWxvY2FsaG9zdDEa
MBgGCSqGSIb3DQEJARYLeHl6QHBkcS5uZXQwHhcNMTUwNTA1MTYzOTIxWhcNMTYw
NTA0MTYzOTIxWjB5MQswCQYDVQQGEwJVUzENMAsGA1UECAwEVXRhaDEXMBUGA1UE
BwwOU2FsdCBMYWtlIENpdHkxEjAQBgNVBAoMCVNhbHRTdGFjazESMBAGA1UEAwwJ
bG9jYWxob3N0MRowGAYJKoZIhvcNAQkBFgt4eXpAcGRxLm5ldDCCASIwDQYJKoZI
hvcNAQEBBQADggEPADCCAQoCggEBAMNvHc8LwpI5/NiwRTWYG34WQ5vau8gkj+8p
5KehXDNmDcCY8QW9xNaCxY6Atg2Dwh5vEacubKRcnQL9SFKYHa4ddtnkISzSkdZN
ImY7ZVQteDIVNJmy7DrZ4RvWTr2ezXYLv8oNkqrKhynt5xIBXZWslWUav1pOp8z8
N+LeXaASVyajqB5TiN8HJR/up9MlSfy/zhtm6x6SIUsEZa+zK7m06/Glrr4WZFOV
LbOwxl36JpjywWTNcrXJd052U/377tUATXpepALBUUOIvWeGF7mrSTZkdhqRZRTe
Jr2+48zIuyMeB+JlY4UpR04pQNqstHimkyjxFfN/TKFqlhYqYjkCAwEAAaOB/DCB
+TASBgNVHRMBAf8ECDAGAQH/AgEAMA4GA1UdDwEB/wQEAwIBBjAdBgNVHQ4EFgQU
WBvk3qjnltkxKtEQxqYn5+KwYWkwgbMGA1UdIwSBqzCBqIAUWBvk3qjnltkxKtEQ
xqYn5+KwYWmhfaR7MHkxCzAJBgNVBAYTAlVTMQ0wCwYDVQQIDARVdGFoMRcwFQYD
VQQHDA5TYWx0IExha2UgQ2l0eTESMBAGA1UECgwJU2FsdFN0YWNrMRIwEAYDVQQD
DAlsb2NhbGhvc3QxGjAYBgkqhkiG9w0BCQEWC3h5ekBwZHEubmV0ghEA1bohQbms
x7PS8+LdxKZLdDANBgkqhkiG9w0BAQsFAAOCAQEALe312Oe8e+VjhnItcjQFuwcP
TaLf3+DTWaQLU1C8H78E75WE9UiRiVCyTpOLt/nONFkIKE275nCLPGCXn5JTZYVB
CxGFTRqnQ+8bdhZA6LYQPXieGikjTy+P2oiKOvPnYsATUXLbZ3ee+zEgBFGbbxNX
Argd3Vahg7Onu3ynsJz9a+hmwVqTX70Ykrrm+b/YtwKPfHeXTMxkX23jc4R7D+ED
VvROFJ27hFPLVaJrsq3EHb8ZkQHmRCzK2sMIPyJb2e0BOKDEZEphNxiIjerDnH7n
xDSMW0jK9FEv/W/sSBwoEolh3Q2e0gfd2vo7bEGvNF4eTqFsoAfeiAjWk65q0Q==
-----END CERTIFICATE-----""",
    "ca_cert_key": """-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQDDbx3PC8KSOfzY
sEU1mBt+FkOb2rvIJI/vKeSnoVwzZg3AmPEFvcTWgsWOgLYNg8IebxGnLmykXJ0C
/UhSmB2uHXbZ5CEs0pHWTSJmO2VULXgyFTSZsuw62eEb1k69ns12C7/KDZKqyocp
7ecSAV2VrJVlGr9aTqfM/Dfi3l2gElcmo6geU4jfByUf7qfTJUn8v84bZusekiFL
BGWvsyu5tOvxpa6+FmRTlS2zsMZd+iaY8sFkzXK1yXdOdlP9++7VAE16XqQCwVFD
iL1nhhe5q0k2ZHYakWUU3ia9vuPMyLsjHgfiZWOFKUdOKUDarLR4ppMo8RXzf0yh
apYWKmI5AgMBAAECggEBAIRqIRRLr4VL7NkUdZAeg2Imy6Apz9mHjE5LYDWDyui4
WNEJzyRIs7lz2U74PmFhyIC+WIOhnNKwPWHtIrdzgYibRg/T1fZ8pXtBv/DshXdH
Z4zneUA6TnyBa1hlF+y6UBOPWl8YWyuFFZd/LXSxoCrtSDu8p7IUYPUuXt9EMsNk
+rQhBZH7iiOx2M7ckLb/gG4tElqh/QQ0Qhuy1e80oMfO9w1XxIThETx0LxQ7Uad5
vk5BQoEI+n/W1tPhXwQ/hndn3ub3pgfpKXIKi7AdySsdxTda4pJLyEwAm9MHrMNm
XvlBdQ+/vq4NSRk2oVVFzJ1yX7bfmrw+6ZqY36oOrUECgYEA4KvnSgeUyG4ouwA9
zTehvFhEWfFyj/y8Tl2OcaIeCvHahN3n6n++Rx25JYrQDhguMGgMHJ0dcZr0Wvvc
EPOAxjam96p2HJMYdpYmrwpguHDu4Jk+fOqgc8Ms541P3+sMej1UQxRaQ6c/g1EM
Pl0g2GdDN4bw60nbEwOXp0qFI1UCgYEA3q+GJDA+gkXcMqKN06VEdjPvm+BpOygx
VvO0cG6njcmVjJKPePIYbjcBHduQRbCGxLmntLOyVg6Ign1+r/LaG7AADYXMdURD
cn4J9ANKuT1lZcs/JRatA1qigwlk8E7vIqq0E/1EiEbRWVE34lgN3dkUOqAXLBfi
9p2aJMOtC1UCgYAKnDOtDFSbbpBf3HAOu/zYXzbDJKLrZ90gukxa03Qlwiw2sCAe
s++xfhbbTgXrVHsB8Df6NfVJAy9dCJ3o8wb21WfnNFalnNC/8PFcvNm6fCLb2oDX
92Ciduos+UB3a6tILpNHI7Prk/9s3Sv92foOHjpPagEAq5k7+aR00xEcjQKBgEhn
r+kCWsDG8ELygcToPqtkVatMO0sF1Y0dLnVENWyvt9V+LfI4XWMwtUc9Bdry+87p
QrNJnlnG3fH31gJlpy9Leajr8T/L01ZdzuStUVWLtfV0MXLgvZ6SkLakjlJoh+6w
rF63gdoBlL5C3zXURaX1mFM7jG1E0wI22lDL4u8FAoGAaeghxaW4V0keS0KpdPvL
kmEhQ5VDAG92fmzQa0wkBoTbsS4kitHj5QTfNJgdCDJbZ7srkMbQR7oVGeZKIDBk
L7lyhyrtHaR/r/fdUNEJAVsQt3Shaa5c5srLh6vzGcBsV7/nQqrEP+GadvGSbHNT
bymYbi0l2pWqQLA2sPoRHNw=
-----END PRIVATE KEY-----""",
    "ca_signed_cert": """
    """,
    "ca_signed_key": """
    """,
    "create_ca": {
        "bits": 2048,
        "CN": "localhost",
        "C": "US",
        "ST": "Utah",
        "L": "Salt Lake City",
        "O": "SaltStack",
        "OU": "Test Unit",
        "emailAddress": "xyz@pdq.net",
        "digest": "sha256",
        "replace": False,
    },
}

# Skip this test case if we don't have access to mock or PyOpenSSL.


@skipIf(NO_PYOPENSSL, "PyOpenSSL must be installed to run these tests.")
class TLSAddTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.tls
    """

    def setup_loader_modules(self):
        return {tls: {}}

    def test_cert_base_path(self):
        """
        Test for retrieving cert base path
        """
        ca_path = "/etc/tls"
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}):
            self.assertEqual(tls.cert_base_path(), ca_path)

    def test_set_ca_cert_path(self):
        """
        Test for setting the cert base path
        """
        ca_path = "/tmp/ca_cert_test_path"
        mock_opt = MagicMock(return_value="/etc/tls")
        ret = {"ca.contextual_cert_base_path": "/tmp/ca_cert_test_path"}
        with patch.dict(tls.__salt__, {"config.option": mock_opt}):
            tls.set_ca_path(ca_path)
            self.assertEqual(tls.__context__, ret)

    def test_ca_exists(self):
        """
        Test to see if ca does not exist
        """
        ca_path = "/tmp/test_tls"
        ca_name = "test_ca"
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
            "os.path.exists", MagicMock(return_value=False)
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertFalse(tls.ca_exists(ca_name))

    def test_ca_exists_true(self):
        """
        Test to see if ca exists
        """
        ca_path = "/tmp/test_tls"
        ca_name = "test_ca"
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
            "os.path.exists", MagicMock(return_value=True)
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertTrue(tls.ca_exists(ca_name))

    def test_get_ca_fail(self):
        """
        Test get_ca failure
        """
        ca_path = "/tmp/test_tls"
        ca_name = "test_ca"
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
            "os.path.exists", MagicMock(return_value=False)
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertRaises(ValueError, tls.get_ca, ca_name)

    def test_get_ca_text(self):
        """
        Test get_ca text
        """
        ca_path = "/tmp/test_tls"
        ca_name = "test_ca"
        mock_opt = MagicMock(return_value=ca_path)
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=_TLS_TEST_DATA["ca_cert"])
        ), patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
            "os.path.exists", MagicMock(return_value=True)
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertEqual(
                tls.get_ca(ca_name, as_text=True), _TLS_TEST_DATA["ca_cert"]
            )

    def test_get_ca(self):
        """
        Test get_ca
        """
        ca_path = "/tmp/test_tls"
        ca_name = "test_ca"
        certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch(
            "os.path.exists", MagicMock(return_value=True)
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertEqual(tls.get_ca(ca_name), certp)

    def test_cert_info(self):
        """
        Test cert info
        """
        self.maxDiff = None
        with patch("os.path.exists", MagicMock(return_value=True)), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            ca_path = "/tmp/test_tls"
            ca_name = "test_ca"
            certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
            ret = {
                "not_after": 1462379961,
                "signature_algorithm": "sha256WithRSAEncryption",
                "extensions": None,
                "fingerprint": (
                    "96:72:B3:0A:1D:34:37:05:75:57:44:7E:08:81:A7:09:"
                    "0C:E1:8F:5F:4D:0C:49:CE:5B:D2:6B:45:D3:4D:FF:31"
                ),
                "serial_number": 284092004844685647925744086791559203700,
                "subject": {
                    "C": "US",
                    "CN": "localhost",
                    "L": "Salt Lake City",
                    "O": "SaltStack",
                    "ST": "Utah",
                    "emailAddress": "xyz@pdq.net",
                },
                "not_before": 1430843961,
                "issuer": {
                    "C": "US",
                    "CN": "localhost",
                    "L": "Salt Lake City",
                    "O": "SaltStack",
                    "ST": "Utah",
                    "emailAddress": "xyz@pdq.net",
                },
            }

            def ignore_extensions(data):
                """
                Ignore extensions pending a resolution of issue 24338
                """
                if "extensions" in data.keys():
                    data["extensions"] = None
                return data

            # older pyopenssl versions don't have extensions or
            # signature_algorithms
            def remove_not_in_result(source, reference):
                if "signature_algorithm" not in reference:
                    del source["signature_algorithm"]
                if "extensions" not in reference:
                    del source["extensions"]

            with patch(
                "salt.utils.files.fopen", mock_open(read_data=_TLS_TEST_DATA["ca_cert"])
            ):
                try:
                    result = ignore_extensions(tls.cert_info(certp))
                except AttributeError as err:
                    # PyOpenSSL version 16.0.0 has an upstream bug in it where a call is made
                    # in OpenSSL/crypto.py in the get_signature_algorithm function referencing
                    # the cert_info attribute, which doesn't exist. This was fixed in subsequent
                    # releases of PyOpenSSL with https://github.com/pyca/pyopenssl/pull/476
                    if (
                        "'_cffi_backend.CDataGCP' object has no attribute 'cert_info'"
                        == str(err)
                    ):
                        log.exception(err)
                        self.skipTest(
                            "Encountered an upstream error with PyOpenSSL: {}".format(
                                err
                            )
                        )
                    if (
                        "'_cffi_backend.CDataGCP' object has no attribute 'object'"
                        == str(err)
                    ):
                        log.exception(err)
                        self.skipTest(
                            "Encountered an upstream error with PyOpenSSL: {}".format(
                                err
                            )
                        )
                    # python-openssl version 0.14, when installed with the "junos-eznc" pip
                    # package, causes an error on this test. Newer versions of PyOpenSSL do not have
                    # this issue. If 0.14 is installed and we hit this error, skip the test.
                    if LooseVersion(OpenSSL.__version__) == LooseVersion("0.14"):
                        log.exception(err)
                        self.skipTest(
                            "Encountered a package conflict. OpenSSL version 0.14"
                            ' cannot be used with the "junos-eznc" pip package on this'
                            " test. Skipping."
                        )
                    result = {}

            remove_not_in_result(ret, result)
            self.assertEqual(result, ret)

    @with_tempdir()
    def test_create_ca(self, ca_path):
        """
        Test creating CA cert
        """
        ca_name = "test_ca"
        certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
        certk = "{0}/{1}/{1}_ca_cert.key".format(ca_path, ca_name)
        ret = 'Created Private Key: "{}." Created CA "{}": "{}."'.format(
            certk, ca_name, certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        with patch.dict(
            tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertEqual(
                tls.create_ca(
                    ca_name, days=365, fixmode=False, **_TLS_TEST_DATA["create_ca"]
                ),
                ret,
            )

    @with_tempdir()
    def test_recreate_ca(self, ca_path):
        """
        Test creating CA cert when one already exists
        """
        ca_name = "test_ca"
        certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
        certk = "{0}/{1}/{1}_ca_cert.key".format(ca_path, ca_name)
        ret = 'Created Private Key: "{}." Created CA "{}": "{}."'.format(
            certk, ca_name, certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        with patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ), patch.dict(
            tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch.dict(
            _TLS_TEST_DATA["create_ca"], {"replace": True}
        ):
            tls.create_ca(ca_name)
            self.assertEqual(
                tls.create_ca(
                    ca_name, days=365, fixmode=False, **_TLS_TEST_DATA["create_ca"]
                ),
                ret,
            )

    @with_tempdir()
    def test_create_csr(self, ca_path):
        """
        Test creating certificate signing request
        """
        ca_name = "test_ca"
        certp = "{}/{}/certs/{}.csr".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        certk = "{}/{}/certs/{}.key".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Private Key: "{}." Created CSR for "{}": "{}."'.format(
            certk, _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        mock_pgt = MagicMock(return_value=False)
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            tls.create_ca(ca_name)
            self.assertEqual(
                tls.create_csr(ca_name, **_TLS_TEST_DATA["create_ca"]), ret
            )

    @with_tempdir()
    def test_recreate_csr(self, ca_path):
        """
        Test creating certificate signing request when one already exists
        """
        ca_name = "test_ca"
        certp = "{}/{}/certs/{}.csr".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        certk = "{}/{}/certs/{}.key".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Private Key: "{}." Created CSR for "{}": "{}."'.format(
            certk, _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        mock_pgt = MagicMock(return_value=False)
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch.dict(
            _TLS_TEST_DATA["create_ca"], {"replace": True}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            tls.create_ca(ca_name)
            tls.create_csr(ca_name)
            self.assertEqual(
                tls.create_csr(ca_name, **_TLS_TEST_DATA["create_ca"]), ret
            )

    @with_tempdir()
    def test_create_self_signed_cert(self, ca_path):
        """
        Test creating self signed certificate
        """
        tls_dir = "test_tls"
        certp = "{}/{}/certs/{}.crt".format(
            ca_path, tls_dir, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        certk = "{}/{}/certs/{}.key".format(
            ca_path, tls_dir, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Private Key: "{}." Created Certificate: "{}."'.format(
            certk, certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertEqual(
                tls.create_self_signed_cert(
                    tls_dir=tls_dir, days=365, **_TLS_TEST_DATA["create_ca"]
                ),
                ret,
            )

    @with_tempdir()
    def test_recreate_self_signed_cert(self, ca_path):
        """
        Test creating self signed certificate when one already exists
        """
        tls_dir = "test_tls"
        certp = "{}/{}/certs/{}.crt".format(
            ca_path, tls_dir, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        certk = "{}/{}/certs/{}.key".format(
            ca_path, tls_dir, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Private Key: "{}." Created Certificate: "{}."'.format(
            certk, certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        with patch.dict(tls.__salt__, {"config.option": mock_opt}), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            self.assertEqual(
                tls.create_self_signed_cert(
                    tls_dir=tls_dir, days=365, **_TLS_TEST_DATA["create_ca"]
                ),
                ret,
            )

    @with_tempdir()
    def test_create_ca_signed_cert(self, ca_path):
        """
        Test signing certificate from request
        """
        ca_name = "test_ca"
        certp = "{}/{}/certs/{}.crt".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Certificate for "{}": "{}"'.format(
            _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        mock_pgt = MagicMock(return_value=False)
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            tls.create_ca(ca_name)
            tls.create_csr(ca_name, **_TLS_TEST_DATA["create_ca"])
            self.assertEqual(
                tls.create_ca_signed_cert(ca_name, _TLS_TEST_DATA["create_ca"]["CN"]),
                ret,
            )

    @with_tempdir()
    def test_recreate_ca_signed_cert(self, ca_path):
        """
        Test signing certificate from request when certificate exists
        """
        ca_name = "test_ca"
        certp = "{}/{}/certs/{}.crt".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Certificate for "{}": "{}"'.format(
            _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        mock_pgt = MagicMock(return_value=False)
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            tls.create_ca(ca_name)
            tls.create_csr(ca_name)
            tls.create_ca_signed_cert(ca_name, _TLS_TEST_DATA["create_ca"]["CN"])
            self.assertEqual(
                tls.create_ca_signed_cert(
                    ca_name, _TLS_TEST_DATA["create_ca"]["CN"], replace=True
                ),
                ret,
            )

    @with_tempdir()
    def test_create_pkcs12(self, ca_path):
        """
        Test creating pkcs12
        """
        ca_name = "test_ca"
        certp = "{}/{}/certs/{}.p12".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created PKCS#12 Certificate for "{}": "{}"'.format(
            _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        mock_pgt = MagicMock(return_value=False)
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            tls.create_ca(ca_name)
            tls.create_csr(ca_name, **_TLS_TEST_DATA["create_ca"])
            tls.create_ca_signed_cert(ca_name, _TLS_TEST_DATA["create_ca"]["CN"])
            self.assertEqual(
                tls.create_pkcs12(
                    ca_name, _TLS_TEST_DATA["create_ca"]["CN"], "password"
                ),
                ret,
            )

    @with_tempdir()
    def test_recreate_pkcs12(self, ca_path):
        """
        Test creating pkcs12 when it already exists
        """
        ca_name = "test_ca"
        certp = "{}/{}/certs/{}.p12".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created PKCS#12 Certificate for "{}": "{}"'.format(
            _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        mock_pgt = MagicMock(return_value=False)
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ), patch.dict(
            tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}
        ), patch.dict(
            _TLS_TEST_DATA["create_ca"], {"replace": True}
        ), patch(
            "salt.modules.tls.maybe_fix_ssl_version", MagicMock(return_value=True)
        ):
            tls.create_ca(ca_name)
            tls.create_csr(ca_name)
            tls.create_ca_signed_cert(ca_name, _TLS_TEST_DATA["create_ca"]["CN"])
            tls.create_pkcs12(ca_name, _TLS_TEST_DATA["create_ca"]["CN"], "password")
            self.assertEqual(
                tls.create_pkcs12(
                    ca_name, _TLS_TEST_DATA["create_ca"]["CN"], "password", replace=True
                ),
                ret,
            )

    def test_pyOpenSSL_version(self):
        """
        Test extension logic with different pyOpenSSL versions
        """
        pillarval = {"csr": {"extendedKeyUsage": "serverAuth"}}
        mock_pgt = MagicMock(return_value=pillarval)
        with patch.dict(
            tls.__dict__,
            {"OpenSSL_version": LooseVersion("0.1.1"), "X509_EXT_ENABLED": False},
        ):
            self.assertEqual(
                tls.__virtual__(),
                (
                    False,
                    "PyOpenSSL version 0.10 or later must be installed "
                    "before this module can be used.",
                ),
            )
            with patch.dict(tls.__salt__, {"pillar.get": mock_pgt}):
                self.assertRaises(AssertionError, tls.get_extensions, "server")
                self.assertRaises(AssertionError, tls.get_extensions, "client")
        with patch.dict(
            tls.__dict__,
            {"OpenSSL_version": LooseVersion("0.14.1"), "X509_EXT_ENABLED": True},
        ):
            self.assertTrue(tls.__virtual__())
            with patch.dict(tls.__salt__, {"pillar.get": mock_pgt}):
                self.assertEqual(tls.get_extensions("server"), pillarval)
                self.assertEqual(tls.get_extensions("client"), pillarval)
        with patch.dict(
            tls.__dict__,
            {"OpenSSL_version": LooseVersion("0.15.1"), "X509_EXT_ENABLED": True},
        ):
            self.assertTrue(tls.__virtual__())
            with patch.dict(tls.__salt__, {"pillar.get": mock_pgt}):
                self.assertEqual(tls.get_extensions("server"), pillarval)
                self.assertEqual(tls.get_extensions("client"), pillarval)

    @with_tempdir()
    def test_pyOpenSSL_version_destructive(self, ca_path):
        """
        Test extension logic with different pyOpenSSL versions
        """
        pillarval = {"csr": {"extendedKeyUsage": "serverAuth"}}
        mock_pgt = MagicMock(return_value=pillarval)
        ca_name = "test_ca"
        certp = "{0}/{1}/{1}_ca_cert.crt".format(ca_path, ca_name)
        certk = "{0}/{1}/{1}_ca_cert.key".format(ca_path, ca_name)
        ret = 'Created Private Key: "{}." Created CA "{}": "{}."'.format(
            certk, ca_name, certp
        )
        mock_opt = MagicMock(return_value=ca_path)
        mock_ret = MagicMock(return_value=0)
        with patch.dict(
            tls.__salt__, {"config.option": mock_opt, "cmd.retcode": mock_ret}
        ):
            with patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}):
                with patch.dict(_TLS_TEST_DATA["create_ca"], {"replace": True}):
                    with patch.dict(
                        tls.__dict__,
                        {
                            "OpenSSL_version": LooseVersion("0.1.1"),
                            "X509_EXT_ENABLED": False,
                        },
                    ):
                        self.assertEqual(
                            tls.create_ca(
                                ca_name,
                                days=365,
                                fixmode=False,
                                **_TLS_TEST_DATA["create_ca"]
                            ),
                            ret,
                        )
                    with patch.dict(
                        tls.__dict__,
                        {
                            "OpenSSL_version": LooseVersion("0.14.1"),
                            "X509_EXT_ENABLED": True,
                        },
                    ):
                        self.assertEqual(
                            tls.create_ca(
                                ca_name,
                                days=365,
                                fixmode=False,
                                **_TLS_TEST_DATA["create_ca"]
                            ),
                            ret,
                        )
                    with patch.dict(
                        tls.__dict__,
                        {
                            "OpenSSL_version": LooseVersion("0.15.1"),
                            "X509_EXT_ENABLED": True,
                        },
                    ):
                        self.assertEqual(
                            tls.create_ca(
                                ca_name,
                                days=365,
                                fixmode=False,
                                **_TLS_TEST_DATA["create_ca"]
                            ),
                            ret,
                        )

        certp = "{}/{}/certs/{}.csr".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        certk = "{}/{}/certs/{}.key".format(
            ca_path, ca_name, _TLS_TEST_DATA["create_ca"]["CN"]
        )
        ret = 'Created Private Key: "{}." Created CSR for "{}": "{}."'.format(
            certk, _TLS_TEST_DATA["create_ca"]["CN"], certp
        )
        with patch.dict(
            tls.__salt__,
            {
                "config.option": mock_opt,
                "cmd.retcode": mock_ret,
                "pillar.get": mock_pgt,
            },
        ):
            with patch.dict(tls.__opts__, {"hash_type": "sha256", "cachedir": ca_path}):
                with patch.dict(
                    _TLS_TEST_DATA["create_ca"],
                    {"subjectAltName": "DNS:foo.bar", "replace": True},
                ):
                    with patch.dict(
                        tls.__dict__,
                        {
                            "OpenSSL_version": LooseVersion("0.1.1"),
                            "X509_EXT_ENABLED": False,
                        },
                    ):
                        tls.create_ca(ca_name)
                        tls.create_csr(ca_name)
                        self.assertRaises(
                            ValueError,
                            tls.create_csr,
                            ca_name,
                            **_TLS_TEST_DATA["create_ca"]
                        )
                    with patch.dict(
                        tls.__dict__,
                        {
                            "OpenSSL_version": LooseVersion("0.14.1"),
                            "X509_EXT_ENABLED": True,
                        },
                    ):
                        tls.create_ca(ca_name)
                        tls.create_csr(ca_name)
                        self.assertEqual(
                            tls.create_csr(ca_name, **_TLS_TEST_DATA["create_ca"]), ret
                        )
                    with patch.dict(
                        tls.__dict__,
                        {
                            "OpenSSL_version": LooseVersion("0.15.1"),
                            "X509_EXT_ENABLED": True,
                        },
                    ):
                        tls.create_ca(ca_name)
                        tls.create_csr(ca_name)
                        self.assertEqual(
                            tls.create_csr(ca_name, **_TLS_TEST_DATA["create_ca"]), ret
                        )

    def test_get_expiration_date(self):
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=_TLS_TEST_DATA["ca_cert"])
        ):
            self.assertEqual(tls.get_expiration_date("/path/to/cert"), "2016-05-04")
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=_TLS_TEST_DATA["ca_cert"])
        ):
            self.assertEqual(
                tls.get_expiration_date("/path/to/cert", date_format="%d/%m/%y"),
                "04/05/16",
            )



"""
Interface to Red Hat tuned-adm module

:maintainer:    Syed Ali <alicsyed@gmail.com>
:maturity:      new
:depends:       cmd.run
:platform:      Linux
"""


import salt.exceptions


def profile(name):
    """
    This state module allows you to modify system tuned parameters

    Example tuned.sls file to set profile to virtual-guest

    tuned:
      tuned:
        - profile
        - name: virtual-guest

    name
        tuned profile name to set the system to

    To see a valid list of states call execution module:
        :py:func:`tuned.list <salt.modules.tuned.list_>`
    """

    # create data-structure to return with default value
    ret = {"name": "", "changes": {}, "result": False, "comment": ""}

    ret[name] = name
    profile = name

    # get the current state of tuned-adm
    current_state = __salt__["tuned.active"]()

    valid_profiles = __salt__["tuned.list"]()

    # check valid profiles, and return error if profile name is not valid
    if profile not in valid_profiles:
        raise salt.exceptions.SaltInvocationError("Invalid Profile Name")

    # if current state is same as requested state, return without doing much
    if profile in current_state:
        ret["result"] = True
        ret["comment"] = "System already in the correct state"
        return ret

    # test mode
    if __opts__["test"] is True:
        ret["comment"] = 'The state of "{}" will be changed.'.format(current_state)
        ret["changes"] = {
            "old": current_state,
            "new": "Profile will be set to {}".format(profile),
        }

        # return None when testing
        ret["result"] = None
        return ret

    # we come to this stage if current state is different that requested state
    # we there have to set the new state request
    new_state = __salt__["tuned.profile"](profile)

    # create the comment data structure
    ret["comment"] = 'The state of "{}" was changed!'.format(profile)

    # fill in the ret data structure
    ret["changes"] = {
        "old": current_state,
        "new": new_state,
    }

    ret["result"] = True

    # return with the dictionary data structure
    return ret


def off(name=None):
    """

    Turns 'tuned' off.
    Example tuned.sls file for turning tuned off:

    tuned:
      tuned.off: []


    To see a valid list of states call execution module:
        :py:func:`tuned.list <salt.modules.tuned.list_>`
    """

    # create data-structure to return with default value
    ret = {"name": "off", "changes": {}, "result": False, "comment": "off"}

    # check the current state of tuned
    current_state = __salt__["tuned.active"]()

    # if profile is already off, then don't do anything
    if current_state == "off":
        ret["result"] = True
        ret["comment"] = "System already in the correct state"
        return ret

    # test mode
    if __opts__["test"] is True:
        ret["comment"] = 'The state of "{}" will be changed.'.format(current_state)
        ret["changes"] = {
            "old": current_state,
            "new": "Profile will be set to off",
        }

        # return None when testing
        ret["result"] = None
        return ret

    # actually execute the off statement
    if __salt__["tuned.off"]():
        ret["result"] = True
        ret["changes"] = {
            "old": current_state,
            "new": "off",
        }
        return ret



import sys

import pytest
from tests.support.case import ModuleCase
from tests.support.unit import skipIf


class SysrcModuleTest(ModuleCase):
    def setUp(self):
        super().setUp()
        ret = self.run_function("cmd.has_exec", ["sysrc"])
        if not ret:
            self.skipTest("sysrc not found")

    @skipIf(not sys.platform.startswith("freebsd"), "FreeBSD specific")
    def test_show(self):
        ret = self.run_function("sysrc.get")
        self.assertIsInstance(
            ret, dict, "sysrc.get returned wrong type, expecting dictionary"
        )
        self.assertIn(
            "/etc/rc.conf", ret, "sysrc.get should have an rc.conf key in it."
        )

    @skipIf(not sys.platform.startswith("freebsd"), "FreeBSD specific")
    @pytest.mark.destructive_test
    def test_set(self):
        ret = self.run_function("sysrc.set", ["test_var", "1"])
        self.assertIsInstance(
            ret, dict, "sysrc.get returned wrong type, expecting dictionary"
        )
        self.assertIn(
            "/etc/rc.conf", ret, "sysrc.set should have an rc.conf key in it."
        )
        self.assertIn(
            "1",
            ret["/etc/rc.conf"]["test_var"],
            "sysrc.set should return the value it set.",
        )
        ret = self.run_function("sysrc.remove", ["test_var"])
        self.assertEqual("test_var removed", ret)

    @skipIf(not sys.platform.startswith("freebsd"), "FreeBSD specific")
    @pytest.mark.destructive_test
    def test_set_bool(self):
        ret = self.run_function("sysrc.set", ["test_var", True])
        self.assertIsInstance(
            ret, dict, "sysrc.get returned wrong type, expecting dictionary"
        )
        self.assertIn(
            "/etc/rc.conf", ret, "sysrc.set should have an rc.conf key in it."
        )
        self.assertIn(
            "YES",
            ret["/etc/rc.conf"]["test_var"],
            "sysrc.set should return the value it set.",
        )
        ret = self.run_function("sysrc.remove", ["test_var"])
        self.assertEqual("test_var removed", ret)



"""
    :codeauthor: Joao Mesquita <jmesquita@sangoma.com>
"""


import datetime
import os
import time

import salt.utils.files
from salt import fileserver
from tests.support.helpers import with_tempdir
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class MapDiffTestCase(TestCase):
    def test_diff_with_diffent_keys(self):
        """
        Test that different maps are indeed reported different
        """
        map1 = {"file1": 1234}
        map2 = {"file2": 1234}
        assert fileserver.diff_mtime_map(map1, map2) is True

    def test_diff_with_diffent_values(self):
        """
        Test that different maps are indeed reported different
        """
        map1 = {"file1": 12345}
        map2 = {"file1": 1234}
        assert fileserver.diff_mtime_map(map1, map2) is True


class VCSBackendWhitelistCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {fileserver: {}}

    def test_whitelist(self):
        opts = {
            "fileserver_backend": ["roots", "git", "s3fs", "hgfs", "svn"],
            "extension_modules": "",
        }
        fs = fileserver.Fileserver(opts)
        assert sorted(fs.servers.whitelist) == sorted(
            ["git", "gitfs", "hg", "hgfs", "svn", "svnfs", "roots", "s3fs"]
        ), fs.servers.whitelist

    @with_tempdir()
    def test_future_file_list_cache_file_ignored(self, cachedir):
        opts = {
            "fileserver_backend": ["roots"],
            "cachedir": cachedir,
            "extension_modules": "",
        }

        back_cachedir = os.path.join(cachedir, "file_lists/roots")
        os.makedirs(os.path.join(back_cachedir))

        # Touch a couple files
        for filename in ("base.p", "foo.txt"):
            with salt.utils.files.fopen(
                os.path.join(back_cachedir, filename), "wb"
            ) as _f:
                if filename == "base.p":
                    _f.write(b"\x80")

        # Set modification time to file list cache file to 1 year in the future
        now = datetime.datetime.utcnow()
        future = now + datetime.timedelta(days=365)
        mod_time = time.mktime(future.timetuple())
        os.utime(os.path.join(back_cachedir, "base.p"), (mod_time, mod_time))

        list_cache = os.path.join(back_cachedir, "base.p")
        w_lock = os.path.join(back_cachedir, ".base.w")
        ret = fileserver.check_file_list_cache(opts, "files", list_cache, w_lock)
        assert (
            ret[1] is True
        ), "Cache file list cache file is not refreshed when future modification time"



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.drbd as drbd
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class DrbdTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.drbd
    """

    def setup_loader_modules(self):
        return {drbd: {}}

    # 'overview' function tests: 1

    def test_overview(self):
        """
        Test if it shows status of the DRBD devices
        """
        ret = {
            "connection state": "True",
            "device": "Stack",
            "fs": "None",
            "local disk state": "UpToDate",
            "local role": "master",
            "minor number": "Salt",
            "mountpoint": "True",
            "partner disk state": "UpToDate",
            "partner role": "minion",
            "percent": "888",
            "remains": "666",
            "total size": "50",
            "used": "50",
        }
        mock = MagicMock(
            return_value=(
                "Salt:Stack True master/minion UpToDate/UpToDate True None 50 50 666 888"
            )
        )
        with patch.dict(drbd.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(drbd.overview(), ret)

        ret = {
            "connection state": "True",
            "device": "Stack",
            "local disk state": "UpToDate",
            "local role": "master",
            "minor number": "Salt",
            "partner disk state": "partner",
            "partner role": "minion",
            "synched": "5050",
            "synchronisation: ": "syncbar",
        }
        mock = MagicMock(
            return_value=(
                "Salt:Stack True master/minion UpToDate/partner syncbar None 50 50"
            )
        )
        with patch.dict(drbd.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(drbd.overview(), ret)

    def test_status(self):
        """
        Test if it shows status of the DRBD resources via drbdadm
        """
        ret = [
            {
                "local role": "Primary",
                "local volumes": [{"disk": "UpToDate"}],
                "peer nodes": [
                    {
                        "peer volumes": [
                            {
                                "done": "96.47",
                                "peer-disk": "Inconsistent",
                                "replication": "SyncSource",
                            }
                        ],
                        "peernode name": "opensuse-node2",
                        "role": "Secondary",
                    }
                ],
                "resource name": "single",
            }
        ]

        mock = MagicMock(
            return_value="""
single role:Primary
  disk:UpToDate
  opensuse-node2 role:Secondary
    replication:SyncSource peer-disk:Inconsistent done:96.47
"""
        )

        with patch.dict(drbd.__salt__, {"cmd.run": mock}):
            try:  # python2
                self.assertItemsEqual(drbd.status(), ret)
            except AttributeError:  # python3
                self.assertCountEqual(drbd.status(), ret)

        ret = [
            {
                "local role": "Primary",
                "local volumes": [
                    {"disk": "UpToDate", "volume": "0"},
                    {"disk": "UpToDate", "volume": "1"},
                ],
                "peer nodes": [
                    {
                        "peer volumes": [
                            {"peer-disk": "UpToDate", "volume": "0"},
                            {"peer-disk": "UpToDate", "volume": "1"},
                        ],
                        "peernode name": "node2",
                        "role": "Secondary",
                    },
                    {
                        "peer volumes": [
                            {"peer-disk": "UpToDate", "volume": "0"},
                            {"peer-disk": "UpToDate", "volume": "1"},
                        ],
                        "peernode name": "node3",
                        "role": "Secondary",
                    },
                ],
                "resource name": "test",
            },
            {
                "local role": "Primary",
                "local volumes": [
                    {"disk": "UpToDate", "volume": "0"},
                    {"disk": "UpToDate", "volume": "1"},
                ],
                "peer nodes": [
                    {
                        "peer volumes": [
                            {"peer-disk": "UpToDate", "volume": "0"},
                            {"peer-disk": "UpToDate", "volume": "1"},
                        ],
                        "peernode name": "node2",
                        "role": "Secondary",
                    },
                    {
                        "peer volumes": [
                            {"peer-disk": "UpToDate", "volume": "0"},
                            {"peer-disk": "UpToDate", "volume": "1"},
                        ],
                        "peernode name": "node3",
                        "role": "Secondary",
                    },
                ],
                "resource name": "res",
            },
        ]

        mock = MagicMock(
            return_value="""
res role:Primary
  volume:0 disk:UpToDate
  volume:1 disk:UpToDate
  node2 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate
  node3 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate

test role:Primary
  volume:0 disk:UpToDate
  volume:1 disk:UpToDate
  node2 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate
  node3 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate

"""
        )
        with patch.dict(drbd.__salt__, {"cmd.run": mock}):
            try:  # python2
                self.assertItemsEqual(drbd.status(), ret)
            except AttributeError:  # python3
                self.assertCountEqual(drbd.status(), ret)
